from numpy import zeros, array, matmul, cross, add
from numpy.linalg import inv, norm, det

#%%
class Plate3D():

    def __init__(self, name, i_node, j_node, m_node, n_node, t, material_name, model, kx_mod=1.0,
                 ky_mod=1.0):
        """
        A rectangular plate element

        Parameters
        ----------
        name : string
            A unique plate name
        i_node : Node3D
            The plate's i-node
        j_node : Node3D
            The plate's j-node
        m_node : Node3D
            The plate's m-node
        n_node : Node3D
            The plate's n-node
        t : number
            Plate thickness
        material_name : string
            The name of the plate material
        kx_mod : number
            Modification factor for stiffness in the plate's local x-direction. Default value is
            1.0, which indicates no stiffness modification (100% stiffness).
        ky_mod : number
            Modification factor for stiffness in the plate's local y-direction. Default value is
            1.0, which indicates no stiffness modification (100% stiffness).
        model : FEModel3D
            The model the plate is a part of
        """

        self.name = name
        self.ID = None
        self.type = 'Rect'

        self.i_node = i_node
        self.j_node = j_node
        self.m_node = m_node
        self.n_node = n_node

        self.t = t
        
        self.kx_mod = kx_mod
        self.ky_mod = ky_mod

        self.pressures = []  # A list of surface pressures [pressure, case='Case 1']

        # Plates need a link to the model they belong to
        self.model = model

        # Get material properties for the plate from the model
        try:
            self.E = self.model.materials[material_name].E
            self.nu = self.model.materials[material_name].nu
        except:
            raise KeyError('Please define the material ' + str(material_name) + ' before assigning it to plates.')
    
    def width(self):
        """
        Returns the width of the plate along its local x-axis
        """
        return self.i_node.distance(self.j_node)

    def height(self):
        """
        Returns the height of the plate along its local y-axis
        """
        return self.i_node.distance(self.n_node)
    
    def Dm(self):
        """
        Returns the plane stress constitutive matrix for orthotropic materials [Dm]
        """

        # Apply the stiffness modification factors for each direction to obtain orthotropic
        # behavior. Stiffness modification factors of 1.0 in each direction (the default) will
        # model isotropic behavior. Orthotropic behavior is limited to the element's local
        # coordinate system.
        Ex = self.E*self.kx_mod
        Ey = self.E*self.ky_mod
        nu_xy = self.nu
        nu_yx = self.nu

        # The shear modulus will be unafected by orthotropic behavior
        # Logan, Appendix C.3, page 750
        G = self.E/(2*(1 + self.nu))

        # Calculate the constitutive matrix [Dm]
        Dm = 1/(1 - nu_xy*nu_yx)*array([[   Ex,    nu_yx*Ex,            0             ],
                                        [nu_xy*Ey,    Ey,               0             ],
                                        [   0,        0,     (1 - nu_xy*nu_yx)*G]])

        # Return the constitutive matrix [Dm]
        return Dm

    def Db(self):
        """
        Returns the bending constitutive matrix for orthotropic materials [Db]
        """

        # Get the plate thickness and other material parameters
        t = self.t
        nu_xy = self.nu
        nu_yx = self.nu
        Ex = self.E*self.kx_mod
        Ey = self.E*self.ky_mod
        G = self.E/(2*(1 + self.nu))

        # Calculate the constitutive matrix [Db]
        Db = t**3/(12*(1 - nu_xy*nu_yx))*array([[   Ex,    nu_yx*Ex, 0],
                                                [nu_xy*Ey,    Ey,    0],
                                                [   0,        0,     G]])

        # Return the constitutive matrix [Db]
        return Db

    def J(self, r, s):
        '''
        Returns the Jacobian matrix for the element
        '''
        
        # Get the local coordinates for the element's nodes
        x1, y1, x2, y2, x3, y3, x4, y4 = 0, 0, self.width(), 0, self.width(), self.height(), 0, self.height()

        # Return the Jacobian matrix
        return 1/4*array([[x1*(s - 1) - x2*(s - 1) + x3*(s + 1) - x4*(s + 1), y1*(s - 1) - y2*(s - 1) + y3*(s + 1) - y4*(s + 1)],
                          [x1*(r - 1) - x2*(r + 1) + x3*(r + 1) - x4*(r - 1), y1*(r - 1) - y2*(r + 1) + y3*(r + 1) - y4*(r - 1)]])
    
    def B_m(self, r, s):

        # Differentiate the interpolation functions
        # Row 1 = interpolation functions differentiated with respect to x
        # Row 2 = interpolation functions differentiated with respect to y
        # Note that the inverse of the Jacobian converts from derivatives with
        # respect to r and s to derivatives with respect to x and y
        dH = matmul(inv(self.J(r, s)), 1/4*array([[s - 1, -s + 1, s + 1, -s - 1],                 
                                                  [r - 1, -r - 1, r + 1, -r + 1]]))

        # Reference 1, Example 5.5 (page 353)
        B_m = array([[dH[0, 0],    0,     dH[0, 1],    0,     dH[0, 2],    0,     dH[0, 3],    0    ],
                     [   0,     dH[1, 0],    0,     dH[1, 1],    0,     dH[1, 2],    0,     dH[1, 3]],
                     [dH[1, 0], dH[0, 0], dH[1, 1], dH[0, 1], dH[1, 2], dH[0, 2], dH[1, 3], dH[0, 3]]])
        
        return B_m
    
    def k(self):
        """
        returns the plate's local stiffness matrix
        """
        return add(self.k_b(), self.k_m())
    
    def k_m(self):
        '''
        Returns the local stiffness matrix for membrane (in-plane) stresses.

        Plane stress is assumed
        '''

        t = self.t
        Dm = self.Dm()

        # Define the gauss point for numerical integration
        gp = 1/3**0.5

        # Get the membrane B matrices for each gauss point
        # Doing this now will save us from doing it twice below
        B1 = self.B_m(-gp, -gp)
        B2 = self.B_m(gp, -gp)
        B3 = self.B_m(gp, gp)
        B4 = self.B_m(-gp, gp)

        # See reference 1 at the bottom of page 353, and reference 2 page 466
        k = t*(matmul(B1.T, matmul(Dm, B1))*det(self.J(-gp, -gp)) +
               matmul(B2.T, matmul(Dm, B2))*det(self.J(gp, -gp)) +
               matmul(B3.T, matmul(Dm, B3))*det(self.J(gp, gp)) +
               matmul(B4.T, matmul(Dm, B4))*det(self.J(-gp, gp)))
        
        k_exp = zeros((24, 24))

        # Step through each term in the unexpanded stiffness matrix
        # i = Unexpanded matrix row
        for i in range(8):

            # j = Unexpanded matrix column
            for j in range(8):
                
                # Find the corresponding term in the expanded stiffness
                # matrix

                # m = Expanded matrix row
                if i in [0, 2, 4, 6]:  # indices associated with displacement in x
                    m = i*3
                if i in [1, 3, 5, 7]:  # indices associated with displacement in y
                    m = i*3 - 2

                # n = Expanded matrix column
                if j in [0, 2, 4, 6]:  # indices associated with displacement in x
                    n = j*3
                if j in [1, 3, 5, 7]:  # indices associated with displacement in y
                    n = j*3 - 2
                
                # Ensure the indices are integers rather than floats
                m, n = round(m), round(n)

                # Add the term from the unexpanded matrix into the expanded matrix
                k_exp[m, n] = k[i, j]
        
        return k_exp
    
    def k_b(self):
        """
        Returns the local stiffness matrix for bending
        """

        b = self.width()/2
        c = self.height()/2

        Ex = self.E
        Ey = self.E
        nu_xy = self.nu
        nu_yx = self.nu
        G = self.E/(2*(1 + self.nu))
        t = self.t

        # Stiffness matrix for plate bending. This matrix was derived using a jupyter notebook. The
        # notebook can be found in the `Derivations`` folder of this project.
        k = t**3/12*array([[(-Ex*nu_yx*b**2*c**2/4 - Ex*c**4 - Ey*nu_xy*b**2*c**2/4 - Ey*b**4 + 7*G*nu_xy*nu_yx*b**2*c**2/5 - 7*G*b**2*c**2/5)/(b**3*c**3*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*c**2/2 - Ey*b**2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (Ex*c**2 + Ey*nu_xy*b**2/2 - G*nu_xy*nu_yx*b**2/5 + G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), (5*Ex*nu_yx*b**2*c**2 + 20*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 - 10*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (Ex*nu_yx*c**2/2 - Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (5*Ex*c**2 - G*nu_xy*nu_yx*b**2 + G*b**2)/(5*b**2*c*(nu_xy*nu_yx - 1)), (-5*Ex*nu_yx*b**2*c**2 + 10*Ex*c**4 - 5*Ey*nu_xy*b**2*c**2 + 10*Ey*b**4 + 28*G*nu_xy*nu_yx*b**2*c**2 - 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (-Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), (5*Ex*nu_yx*b**2*c**2 - 10*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 + 20*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (-5*Ey*b**2 + G*nu_xy*nu_yx*c**2 - G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), (Ex*c**2/2 - Ey*nu_xy*b**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1))],
                           [(-Ey*nu_xy*c**2/2 - Ey*b**2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 4*(-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), Ey*nu_xy/(nu_xy*nu_yx - 1), (Ey*nu_xy*c**2/2 - Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - 4*G*nu_xy*nu_yx*c**2 + 4*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0],
                           [(Ex*nu_yx*b**2/2 + Ex*c**2 - G*nu_xy*nu_yx*b**2/5 + G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), Ex*nu_yx/(nu_xy*nu_yx - 1), 4*(-5*Ex*c**2 + 2*G*nu_xy*nu_yx*b**2 - 2*G*b**2)/(15*b*c*(nu_xy*nu_yx - 1)), (-5*Ex*c**2 + G*nu_xy*nu_yx*b**2 - G*b**2)/(5*b**2*c*(nu_xy*nu_yx - 1)), 0, 2*(-5*Ex*c**2 - G*nu_xy*nu_yx*b**2 + G*b**2)/(15*b*c*(nu_xy*nu_yx - 1)), -(Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, (-5*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*b**2/2 + Ex*c**2/2 + G*nu_xy*nu_yx*b**2/5 - G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, -(10*Ex*c**2 + 8*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1))],
                           [(5*Ex*nu_yx*b**2*c**2 + 20*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 - 10*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (Ex*nu_yx*c**2/2 - Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-5*Ex*c**2 + G*nu_xy*nu_yx*b**2 - G*b**2)/(5*b**2*c*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*b**2*c**2/4 - Ex*c**4 - Ey*nu_xy*b**2*c**2/4 - Ey*b**4 + 7*G*nu_xy*nu_yx*b**2*c**2/5 - 7*G*b**2*c**2/5)/(b**3*c**3*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*c**2/2 - Ey*b**2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-Ex*c**2 - Ey*nu_xy*b**2/2 + G*nu_xy*nu_yx*b**2/5 - G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), (5*Ex*nu_yx*b**2*c**2 - 10*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 + 20*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (-5*Ey*b**2 + G*nu_xy*nu_yx*c**2 - G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), (-Ex*c**2/2 + Ey*nu_xy*b**2/2 - G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), (-5*Ex*nu_yx*b**2*c**2 + 10*Ex*c**4 - 5*Ey*nu_xy*b**2*c**2 + 10*Ey*b**4 + 28*G*nu_xy*nu_yx*b**2*c**2 - 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (-Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), -(Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1))],
                           [(Ey*nu_xy*c**2/2 - Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - 4*G*nu_xy*nu_yx*c**2 + 4*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (-Ey*nu_xy*c**2/2 - Ey*b**2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 4*(-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), -Ey*nu_xy/(nu_xy*nu_yx - 1), (5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0],
                           [(5*Ex*c**2 - G*nu_xy*nu_yx*b**2 + G*b**2)/(5*b**2*c*(nu_xy*nu_yx - 1)), 0, 2*(-5*Ex*c**2 - G*nu_xy*nu_yx*b**2 + G*b**2)/(15*b*c*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*b**2/2 - Ex*c**2 + G*nu_xy*nu_yx*b**2/5 - G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), -Ex*nu_yx/(nu_xy*nu_yx - 1), 4*(-5*Ex*c**2 + 2*G*nu_xy*nu_yx*b**2 - 2*G*b**2)/(15*b*c*(nu_xy*nu_yx - 1)), (Ex*nu_yx*b**2/2 - Ex*c**2/2 - G*nu_xy*nu_yx*b**2/5 + G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, -(10*Ex*c**2 + 8*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), (Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, (-5*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1))],
                           [(-5*Ex*nu_yx*b**2*c**2 + 10*Ex*c**4 - 5*Ey*nu_xy*b**2*c**2 + 10*Ey*b**4 + 28*G*nu_xy*nu_yx*b**2*c**2 - 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), -(Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), (5*Ex*nu_yx*b**2*c**2 - 10*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 + 20*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), (-5*Ex*c**2 - 25*Ey*nu_xy*b**2 + 2*b**2*(15*Ey*nu_xy - G*nu_xy*nu_yx + G))/(10*b**2*c*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*b**2*c**2/4 - Ex*c**4 - Ey*nu_xy*b**2*c**2/4 - Ey*b**4 + 7*G*nu_xy*nu_yx*b**2*c**2/5 - 7*G*b**2*c**2/5)/(b**3*c**3*(nu_xy*nu_yx - 1)), (Ex*nu_yx*c**2/2 + Ey*b**2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-Ex*c**2 - Ey*nu_xy*b**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), (5*Ex*nu_yx*b**2*c**2 + 20*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 - 10*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*c**2/2 + Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-Ex*c**2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1))],
                           [(-Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (-5*Ey*b**2 + G*nu_xy*nu_yx*c**2 - G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (Ey*nu_xy*c**2/2 + Ey*b**2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 4*(-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), Ey*nu_xy/(nu_xy*nu_yx - 1), (-Ey*nu_xy*c**2/2 + Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - 4*G*nu_xy*nu_yx*c**2 + 4*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0],
                           [(Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, (-5*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), (Ex*nu_yx*b**2/2 - Ex*c**2/2 - G*nu_xy*nu_yx*b**2/5 + G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, -(10*Ex*c**2 + 8*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*b**2/2 - Ex*c**2 + G*nu_xy*nu_yx*b**2/5 - G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), Ex*nu_yx/(nu_xy*nu_yx - 1), 4*(-5*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), (Ex*c**2 - G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, -(10*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1))],
                           [(5*Ex*nu_yx*b**2*c**2 - 10*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 + 20*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), (5*Ex*c**2 + 25*Ey*nu_xy*b**2 - 2*b**2*(15*Ey*nu_xy - G*nu_xy*nu_yx + G))/(10*b**2*c*(nu_xy*nu_yx - 1)), (-5*Ex*nu_yx*b**2*c**2 + 10*Ex*c**4 - 5*Ey*nu_xy*b**2*c**2 + 10*Ey*b**4 + 28*G*nu_xy*nu_yx*b**2*c**2 - 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), (5*Ex*nu_yx*b**2*c**2 + 20*Ex*c**4 + 5*Ey*nu_xy*b**2*c**2 - 10*Ey*b**4 - 28*G*nu_xy*nu_yx*b**2*c**2 + 28*G*b**2*c**2)/(20*b**3*c**3*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*c**2/2 + Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (Ex*c**2 - G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), (-Ex*nu_yx*b**2*c**2/4 - Ex*c**4 - Ey*nu_xy*b**2*c**2/4 - Ey*b**4 + 7*G*nu_xy*nu_yx*b**2*c**2/5 - 7*G*b**2*c**2/5)/(b**3*c**3*(nu_xy*nu_yx - 1)), (Ex*nu_yx*c**2/2 + Ey*b**2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (Ex*c**2 + Ey*nu_xy*b**2/2 - G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1))],
                           [(-5*Ey*b**2 + G*nu_xy*nu_yx*c**2 - G*c**2)/(5*b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - G*nu_xy*nu_yx*c**2 + G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (-Ey*b**2/2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), (-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (-Ey*nu_xy*c**2/2 + Ey*b**2/2 + G*nu_xy*nu_yx*c**2/5 - G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 2*(-5*Ey*b**2 - 4*G*nu_xy*nu_yx*c**2 + 4*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), 0, (Ey*nu_xy*c**2/2 + Ey*b**2 - G*nu_xy*nu_yx*c**2/5 + G*c**2/5)/(b*c**2*(nu_xy*nu_yx - 1)), 4*(-5*Ey*b**2 + 2*G*nu_xy*nu_yx*c**2 - 2*G*c**2)/(15*b*c*(nu_xy*nu_yx - 1)), -Ey*nu_xy/(nu_xy*nu_yx - 1)],
                           [(-Ex*nu_yx*b**2/2 + Ex*c**2/2 + G*nu_xy*nu_yx*b**2/5 - G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, -(10*Ex*c**2 + 8*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), -(Ex*c**2/2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, (-5*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), (-Ex*c**2 + G*b**2*(nu_xy*nu_yx - 1)/5)/(b**2*c*(nu_xy*nu_yx - 1)), 0, -(10*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1)), (Ex*nu_yx*b**2/2 + Ex*c**2 - G*nu_xy*nu_yx*b**2/5 + G*b**2/5)/(b**2*c*(nu_xy*nu_yx - 1)), -Ex*nu_yx/(nu_xy*nu_yx - 1), 4*(-5*Ex*c**2 + 2*G*b**2*(nu_xy*nu_yx - 1))/(15*b*c*(nu_xy*nu_yx - 1))]])
        
        # Calculate the stiffness of a weak spring for the drilling degree of freedom (rotation
        # about local z). We'll set the weak spring to be 1000 times weaker than any of the other
        # rotational stiffnesses in the matrix.
        k_rz = min(abs(k[1, 1]), abs(k[2, 2]), abs(k[4, 4]), abs(k[5, 5]),
                   abs(k[7, 7]), abs(k[8, 8]), abs(k[10, 10]), abs(k[11, 11])
                   )/1000

        # The matrix currently only holds terms related to bending action. We need to expand it to
        # with placeholders for all the degrees of freedom so it can be directly added to the
        # membrane stiffness matrix later on.

        # Initialize the expanded stiffness matrix to all zeros
        k_exp = zeros((24, 24))

        # Step through each term in the unexpanded stiffness matrix

        # i = Unexpanded matrix row
        for i in range(12):

            # j = Unexpanded matrix column
            for j in range(12):
                
                # Find the corresponding term in the expanded stiffness
                # matrix

                # m = Expanded matrix row
                if i in [0, 3, 6, 9]:  # indices associated with deflection in z
                    m = 2*i + 2
                if i in [1, 4, 7, 10]:  # indices associated with rotation about x
                    m = 2*i + 1
                if i in [2, 5, 8, 11]:  # indices associated with rotation about y
                    m = 2*i

                # n = Expanded matrix column
                if j in [0, 3, 6, 9]:  # indices associated with deflection in z
                    n = 2*j + 2
                if j in [1, 4, 7, 10]:  # indices associated with rotation about x
                    n = 2*j + 1
                if j in [2, 5, 8, 11]:  # indices associated with rotation about y
                    n = 2*j
                
                # Ensure the indices are integers rather than floats
                m, n = round(m), round(n)

                # Add the term from the unexpanded matrix into the expanded
                # matrix
                k_exp[m, n] = k[i, j]
        
        # Add the drilling degree of freedom's weak spring
        k_exp[5, 5] = k_rz
        k_exp[11, 11] = k_rz
        k_exp[17, 17] = k_rz
        k_exp[23, 23] = k_rz
        
        # Return the local stiffness matrix
        return k_exp

    def f(self, combo_name='Combo 1'):
        """
        Returns the plate's local end force vector
        """
        
        # Calculate and return the plate's local end force vector
        return add(matmul(self.k(), self.d(combo_name)), self.fer(combo_name))

    def fer(self, combo_name='Combo 1'):
        """
        Returns the rectangle's local fixed end reaction vector.

        Parameters
        ----------
        combo_name : string
            The name of the load combination to get the load vector for.
        """
        
        # Initialize the fixed end reaction vector
        fer = zeros((12, 1))

        # Get the requested load combination
        combo = self.model.load_combos[combo_name]

        # Initialize the element's surface pressure to zero
        p = 0
        
        # Loop through each load case and factor in the load combination 
        for case, factor in combo.factors.items():

            # Sum the pressures
            for pressure in self.pressures:

                # Check if the current pressure corresponds to the current load case
                if pressure[1] == case:

                    # Sum the pressures multiplied by their load factors
                    p += factor*pressure[0]
        
        b = self.width()/2
        c = self.height()/2
        
        fer = -4*p*c*b*array([[1/4], [c/12], [-b/12], [1/4], [c/12], [b/12], [1/4], [-c/12], [b/12], [1/4], [-c/12], [-b/12]])

        # At this point `fer` only contains terms for the degrees of freedom
        # associated with membrane action. Expand `fer` to include zero terms for
        # the degrees of freedom related to bending action. This will allow
        # the bending and membrane vectors to be summed directly
        # later on. `numpy` has an `insert` function that can be used to
        # insert rows and columns of zeros one at a time, but it is very slow
        # as it makes a temporary copy of the vector term by term each time
        # it's called. The algorithm used here accomplishes the same thing
        # much faster. Terms are copied only once.

        # Initialize the expanded vector to all zeros
        fer_exp = zeros((24, 1))

        # Step through each term in the unexpanded vector
        # i = Unexpanded vector row
        for i in range(12):
                
            # Find the corresponding term in the expanded vector

            # m = Expanded vector row
            if i in [0, 3, 6, 9]:   # indices associated with deflection in z
                m = 2*i + 2
            if i in [1, 4, 7, 10]:  # indices associated with rotation about x
                m = 2*i + 1
            if i in [2, 5, 8, 11]:  # indices associated with rotation about y
                m = 2*i
                
            # Ensure the index is an integer rather than a float
            m = round(m)

            # Add the term from the unexpanded vector into the expanded vector
            fer_exp[m, 0] = fer[i, 0]

        return fer_exp
        
    def d(self, combo_name='Combo 1'):
       """
       Returns the plate's local displacement vector
       """

       # Calculate and return the local displacement vector
       return matmul(self.T(), self.D(combo_name))

    def F(self, combo_name='Combo 1'):
        """
        Returns the plate's global nodal force vector
        """

        # Calculate and return the global force vector
        return matmul(inv(self.T()), self.f(combo_name))

    def D(self, combo_name='Combo 1'):
        """
        Returns the plate's global displacement vector for the given load combination.
        """
        
        # Initialize the displacement vector
        D = zeros((24, 1))
        
        # Read in the global displacements from the nodes
        D[0, 0] = self.i_node.DX[combo_name]
        D[1, 0] = self.i_node.DY[combo_name]
        D[2, 0] = self.i_node.DZ[combo_name]
        D[3, 0] = self.i_node.RX[combo_name]
        D[4, 0] = self.i_node.RY[combo_name]
        D[5, 0] = self.i_node.RZ[combo_name]

        D[6, 0] = self.j_node.DX[combo_name]
        D[7, 0] = self.j_node.DY[combo_name]
        D[8, 0] = self.j_node.DZ[combo_name]
        D[9, 0] = self.j_node.RX[combo_name]
        D[10, 0] = self.j_node.RY[combo_name]
        D[11, 0] = self.j_node.RZ[combo_name]

        D[12, 0] = self.m_node.DX[combo_name]
        D[13, 0] = self.m_node.DY[combo_name]
        D[14, 0] = self.m_node.DZ[combo_name]
        D[15, 0] = self.m_node.RX[combo_name]
        D[16, 0] = self.m_node.RY[combo_name]
        D[17, 0] = self.m_node.RZ[combo_name]

        D[18, 0] = self.n_node.DX[combo_name]
        D[19, 0] = self.n_node.DY[combo_name]
        D[20, 0] = self.n_node.DZ[combo_name]
        D[21, 0] = self.n_node.RX[combo_name]
        D[22, 0] = self.n_node.RY[combo_name]
        D[23, 0] = self.n_node.RZ[combo_name]
        
        # Return the global displacement vector
        return D
 
    def T(self):
        """
        Returns the plate's transformation matrix
        """

        # Calculate the direction cosines for the local x-axis
        # The local x-axis will run from the i-node to the j-node
        xi = self.i_node.X
        xj = self.j_node.X
        yi = self.i_node.Y
        yj = self.j_node.Y
        zi = self.i_node.Z
        zj = self.j_node.Z
        x = [(xj - xi), (yj - yi), (zj - zi)]
        x = x/norm(x)
        
        # The local y-axis will be in the plane of the plate
        # Find a vector in the plate's local xy plane
        xn = self.n_node.X
        yn = self.n_node.Y
        zn = self.n_node.Z
        xy = [xn - xi, yn - yi, zn - zi]

        # Find a vector perpendicular to the plate surface to get the orientation of the local z-axis
        z = cross(x, xy)
        
        # Divide the vector by its magnitude to produce a unit z-vector of direction cosines
        z = z/norm(z)

        # Calculate the local y-axis as a vector perpendicular to the local z and x-axes
        y = cross(z, x)
        
        # Divide the z-vector by its magnitude to produce a unit vector of direction cosines
        y = y/norm(y)

        # Create the direction cosines matrix
        dirCos = array([x, y, z])
        
        # Build the transformation matrix
        transMatrix = zeros((24, 24))
        transMatrix[0:3, 0:3] = dirCos
        transMatrix[3:6, 3:6] = dirCos
        transMatrix[6:9, 6:9] = dirCos
        transMatrix[9:12, 9:12] = dirCos
        transMatrix[12:15, 12:15] = dirCos
        transMatrix[15:18, 15:18] = dirCos
        transMatrix[18:21, 18:21] = dirCos
        transMatrix[21:24, 21:24] = dirCos
        
        return transMatrix

    def K(self):
        """
        Returns the plate's global stiffness matrix
        """

        # Calculate and return the stiffness matrix in global coordinates
        return matmul(matmul(inv(self.T()), self.k()), self.T())

    def FER(self, combo_name='Combo 1'):
        """
        Returns the global fixed end reaction vector.

        Parameters
        ----------
        combo_name : string
            The name of the load combination to calculate the fixed end
            reaction vector for (not the load combination itself).
        """
        
        # Calculate and return the fixed end reaction vector
        return matmul(inv(self.T()), self.fer(combo_name))

    def _C(self):
        """
        Returns the plate's displacement coefficient matrix [C]
        """

        # Find the local x and y coordinates at each node
        xi = 0
        yi = 0
        xj = self.width()
        yj = 0
        xm = xj
        ym = self.height()
        xn = 0
        yn = ym

        # Calculate the [C] coefficient matrix
        C = array([[1, xi, yi, xi**2, xi*yi, yi**2, xi**3, xi**2*yi, xi*yi**2, yi**3, xi**3*yi, xi*yi**3],
                    [0, 0, 1, 0, xi, 2*yi, 0, xi**2, 2*xi*yi, 3*yi**2, xi**3, 3*xi*yi**2],
                    [0, -1, 0, -2*xi, -yi, 0, -3*xi**2, -2*xi*yi, -yi**2, 0, -3*xi**2*yi, -yi**3],
                    
                    [1, xj, yj, xj**2, xj*yj, yj**2, xj**3, xj**2*yj, xj*yj**2, yj**3, xj**3*yj, xj*yj**3],
                    [0, 0, 1, 0, xj, 2*yj, 0, xj**2, 2*xj*yj, 3*yj**2, xj**3, 3*xj*yj**2],
                    [0, -1, 0, -2*xj, -yj, 0, -3*xj**2, -2*xj*yj, -yj**2, 0, -3*xj**2*yj, -yj**3],

                    [1, xm, ym, xm**2, xm*ym, ym**2, xm**3, xm**2*ym, xm*ym**2, ym**3, xm**3*ym, xm*ym**3],
                    [0, 0, 1, 0, xm, 2*ym, 0, xm**2, 2*xm*ym, 3*ym**2, xm**3, 3*xm*ym**2],
                    [0, -1, 0, -2*xm, -ym, 0, -3*xm**2, -2*xm*ym, -ym**2, 0, -3*xm**2*ym, -ym**3],

                    [1, xn, yn, xn**2, xn*yn, yn**2, xn**3, xn**2*yn, xn*yn**2, yn**3, xn**3*yn, xn*yn**3],
                    [0, 0, 1, 0, xn, 2*yn, 0, xn**2, 2*xn*yn, 3*yn**2, xn**3, 3*xn*yn**2],
                    [0, -1, 0, -2*xn, -yn, 0, -3*xn**2, -2*xn*yn, -yn**2, 0, -3*xn**2*yn, -yn**3]])
        
        # Return the coefficient matrix
        return C

    def _Q(self, x, y):
        """
        Calculates and returns the plate curvature coefficient matrix [Q] at a given point (x, y)
        in the plate's local system.
        """

        # Calculate the [Q] coefficient matrix
        Q =  array([[0, 0, 0, -2, 0, 0, -6*x, -2*y, 0, 0, -6*x*y, 0],
                    [0, 0, 0, 0, 0, -2, 0, 0, -2*x, -6*y, 0, -6*x*y],
                    [0, 0, 0, 0, -2, 0, 0, -4*x, -4*y, 0, -6*x**2, -6*y**2]])
 
        # Return the [Q] coefficient matrix
        return Q

    def _a(self, combo_name='Combo 1'):
        """
        Returns the vector of plate bending constants for the displacement function.

        Parameters
        ----------
        combo_name : string
            The name of the load combination to get the vector for
        """

        # Get the plate's local displacement vector
        # Slice out terms not related to plate bending
        d = self.d(combo_name)[[2, 3, 4, 8, 9, 10, 14, 15, 16, 20, 21, 22], :]

        # Return the plate bending constants
        return inv(self._C()) @ d

    def moment(self, x, y, local=True, combo_name='Combo 1'):
        """
        Returns the internal moments (Mx, My, and Mxy) at any point (x, y) in the plate's local
        coordinate system

        Parameters
        ----------
        x : number
            The x-coordinate in the plate's local coordinate system.
        y : number
            The y-coordinate in the plate's local coordinate system.
        combo_name : string
            The name of the load combination to evaluate. The default is 'Combo 1'.

        """
        
        # Calculate and return internal moments
        # A negative sign will be applied to change the sign convention to match that of
        # Pynite's quadrilateral elements.
        return -self.Db() @ self._Q(x, y) @ self._a(combo_name)
 
    def shear(self, x, y, local=True, combo_name='Combo 1'):
        """
        Returns the internal shears (Qx and Qy) at any point (x, y) in the plate's local
        coordinate system

        Parameters
        ----------
        x : number
            The x-coordinate in the plate's local coordinate system.
        y : number
            The y-coordinate in the plate's local coordinate system.
        combo_name : string
            The name of the load combination to evaluate. The default is 'Combo 1'.

        """

        # Store matrices into local variables for quicker access
        Db = self.Db()
        a = self._a(combo_name)

        # Calculate the derivatives of the plate moments needed to compute shears
        dMx_dx = (Db @ array([[0, 0, 0, 0, 0, 0, -6, 0, 0, 0, -6*y, 0],
                             [0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, -6*y],
                             [0, 0, 0, 0, 0, 0, 0, -4, 0, 0, -12*x, 0]]) @ a)[0]

        dMxy_dy = (Db @ array([[0, 0, 0, 0, 0, 0, 0, -2, 0, 0, -6*x, 0],
                              [0, 0, 0, 0, 0, 0, 0, 0, 0, -6, 0, -6*x],
                              [0, 0, 0, 0, 0, 0, 0, 0, -4, 0, 0, -12*y]]) @ a)[2]
        
        dMy_dy = (Db @ array([[0, 0, 0, 0, 0, 0, 0, -2, 0, 0, -6*x, 0],
                             [0, 0, 0, 0, 0, 0, 0, 0, 0, -6, 0, -6*x],
                             [0, 0, 0, 0, 0, 0, 0, 0, -4, 0, 0, -12*y]]) @ a)[1]

        dMxy_dx = (Db @ array([[0, 0, 0, 0, 0, 0, -6, 0, 0, 0, -6*y, 0],
                              [0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, -6*y],
                              [0, 0, 0, 0, 0, 0, 0, -4, 0, 0, -12*x, 0]]) @ a)[2]
        
        # Calculate internal shears
        Qx = (dMx_dx + dMxy_dy)[0]
        Qy = (dMy_dy + dMxy_dx)[0]

        # Return internal shears
        return array([[Qx], 
                      [Qy]])

    def membrane(self, x, y, local=True, combo_name='Combo 1'):
        
        # Convert the (x, y) coordinates to (r, x) coordinates
        r = -1 + 2*x/self.width()
        s = -1 + 2*y/self.height()

        # Get the plate's local displacement vector
        # Slice out terms not related to membrane forces
        d = self.d(combo_name)[[0, 1, 6, 7, 12, 13, 18, 19], :]

        # Define the gauss point used for numerical integration
        gp = 1/3**0.5

        # Define extrapolated r and s points
        r_ex = r/gp
        s_ex = s/gp

        # Define the interpolation functions
        H = 1/4*array([(1 - r_ex)*(1 - s_ex), (1 + r_ex)*(1 - s_ex), (1 + r_ex)*(1 + s_ex), (1 - r_ex)*(1 + s_ex)])

        # Get the stress-strain matrix
        Dm = self.Dm()
        
        # Calculate the internal stresses [Sx, Sy, Txy] at each gauss point
        s1 = matmul(Dm, matmul(self.B_m(-gp, -gp), d))
        s2 = matmul(Dm, matmul(self.B_m(gp, -gp), d))
        s3 = matmul(Dm, matmul(self.B_m(gp, gp), d))
        s4 = matmul(Dm, matmul(self.B_m(-gp, gp), d))

        # Extrapolate to get the value at the requested location
        Sx = H[0]*s1[0] + H[1]*s2[0] + H[2]*s3[0] + H[3]*s4[0]
        Sy = H[0]*s1[1] + H[1]*s2[1] + H[2]*s3[1] + H[3]*s4[1]
        Txy = H[0]*s1[2] + H[1]*s2[2] + H[2]*s3[2] + H[3]*s4[2]

        return array([Sx,
                      Sy,
                      Txy])



# -*- coding: utf-8 -*-
"""
Created on Mon Nov  6 20:52:31 2017

@author: D. Craig Brinck, SE
"""
from numpy import full, array

# %%
# A mathematically continuous beam segment
class BeamSegZ():
    """
    A mathematically continuous beam segment
    
    Properties
    ----------
    x1 : number
      The starting location of the segment relative to the start of the beam
    x2 : number
      The ending location of the segment relative to the start of the beam
    w1 : number
      The distributed load magnitude at the start of the segment
    w2 : number
      The distributed load magnitude at the end of the segment
    p1 : number
      The distributed axial load magnitude at the start of the segment
    p2 : number
      The distributed axial load magnitude at the end of the segment
    V1 : number
      The internal shear force at the start of the segment
    M1 : number
      The internal moment at the start of the segment
    P1 : number
      The internal axial force at the start of the segment
    T1 : number
      Torsional moment at start of segment
    theta1: number
      The slope (radians) at the start of the segment
    delta1: number
      The transverse displacement at the start of the segment
    delta_x1 : number
      The axial displacement at the start of the segment
    EI : number
      The flexural stiffness of the segment
    EA : number
      The axial stiffness of the segment
    
    Methods
    -------
    __init__()
      Constructor
    Length()
      Returns the length of the segment
    Shear(x)
      Returns the shear force at a location on the segment
    
    Notes
    -----
    Any unit system may be used as long as the units are consistent with each other
    
    """

    def __init__(self):
        """
        Constructor
        """
        
        self.x1 = None # Start location of beam segment (relative to start of beam)
        self.x2 = None # End location of beam segment (relative to start of beam)
        self.w1 = None # Linear distributed transverse load at start of segment
        self.w2 = None # Linear distributed transverse load at end of segment
        self.p1 = None # Linear distributed axial load at start of segment
        self.p2 = None # Linear distributed axial load at end of segment
        self.V1 = None # Internal shear force at start of segment
        self.M1 = None # Internal moment at start of segment
        self.P1 = None # Internal axial force at start of segment
        self.T1 = None # Torsional moment at start of segment
        self.theta1 = None # Slope at start of beam segment
        self.delta1 = None # Displacement at start of beam segment
        self.delta_x1 = None # Axial displacement at start of beam segment
        self.EI = None # Flexural stiffness of the beam segment
        self.EA = None # Axial stiffness of the beam segment

    # Returns the length of the segment
    def Length(self):
        """
        Returns the length of the segment
        """
        
        return self.x2 - self.x1

    # Returns the shear force at a location 'x' on the segment
    def Shear(self, x):
        
        V1 = self.V1
        w1 = self.w1
        w2 = self.w2
        L = self.Length()
        
        return V1 + w1*x + x**2*(-w1 + w2)/(2*L)

    # Returns the moment at a location on the segment
    def moment(self, x, P_delta=False):

        V1 = self.V1
        M1 = self.M1
        P1 = self.P1
        w1 = self.w1
        w2 = self.w2
        L = self.Length()
        
        M = M1 - V1*x - w1*x**2/2 - x**3*(-w1 + w2)/(6*L)

        # Include the P-little-delta moment if a P-Delta analysis was run
        if P_delta == True:
            delta_1 = self.delta1
            delta_x = self.deflection(x)
            M += P1*(delta_x - delta_1)
        
        # Return the computed moment
        return M

    # Returns the axial force at a location on the segment
    def axial(self, x):
        
        P1 = self.P1
        p1 = self.p1
        p2 = self.p2
        L = self.Length()
        
        return P1 + (p2 - p1)/(2*L)*x**2 + p1*x

    def Torsion(self, x = 0):
        """
        Returns the torsional moment in the segment.
        """
        
        # The torsional moment is constant across the segment
        # This can be updated in the future for distributed torsional forces

        #As the return value is not calculated as a function of x (for now), we need to check
        #whether x is an array, and if so, return a results array of the same length
        if isinstance(x, (int, float)):
            return self.T1
        else:
            return full(len(x), self.T1)
    
    def slope(self, x, P_delta=False):
        """Returns the slope of the elastic curve at any point `x` along the segment.

        :param x: Location (relative to start of segment) where slope is to be calculated.
        :type x: float
        :param P_delta: Indicates whether P-little-delta effects should be included in the calculation. Generally only used when a P-Delta analysis has been run. Defaults to False.
        :type P_delta: bool, optional
        :return: The slope of the elastic curve (radians) at location `x`.
        :rtype: float
        """
        
        V1 = self.V1
        M1 = self.M1
        P1 = self.P1
        w1 = self.w1
        w2 = self.w2
        theta_1 = self.theta1
        L = self.Length()
        EI = self.EI

        theta_x = theta_1 - (-V1*x**2/2 - w1*x**3/6 + x*M1 + x**4*(w1 - w2)/(24*L))/EI
        theta_x = theta_1 - (-V1*x**2/2 - w1*x**3/6 + x*M1 + x**4*(w1 - w2)/(24*L))/EI

        if P_delta == True:
            delta_1 = self.delta1
            delta_x = self.deflection(x, P_delta)
            theta_x -= x*(-P1*delta_1 + P1*delta_x)/EI
        
        # TODO: This is an old equation left for reference. Delete it after the new equation has been proved over time.
        # return theta_1 - (M1*x - V1*x**2/2 - w1*x**3/6 + x**4*(w1 - w2)/(24*L))/(EI)
        
        # Return the calculated slope
        return theta_x

    # Returns the deflection at a location on the segment
    def deflection(self, x, P_delta=False):
        
        V1 = self.V1
        M1 = self.M1
        P1 = self.P1
        w1 = self.w1
        w2 = self.w2
        theta_1 = self.theta1
        delta_1 = self.delta1
        d_delta = 1
        L = self.Length()
        EI = self.EI
        
        # Iteration is required to calculate P-little-delta effects
        if P_delta == True:

            # Initialize the deflection at `x` to match the deflection at the start of the segment
            delta_x = delta_1

            # Iterate until we reach a deflection convergence of 1%
            while d_delta > 0.01: 
                
                # Save the deflection value from the last iteration
                delta_last = delta_x

                # Compute the deflection
                # delta_x = delta1 + theta1*x + V1*x**3/(6*EI) + w1*x**4/(24*EI) + x**2*(-M1 + P1*delta1 - P1*delta_x)/(2*EI) + x**5*(-w1 + w2)/(120*EI*L)
                delta_x = delta_1 + theta_1*x + V1*x**3/(6*EI) + w1*x**4/(24*EI) + x**2*(-M1 + P1*delta_1 - P1*delta_x)/(2*EI) + x**5*(-w1 + w2)/(120*EI*L)
                
                # Check the change in deflection between iterations
                if delta_last != 0:
                    d_delta = abs(delta_x/delta_last - 1)
                else:
                    # Members with no relative deflection after at least one iteration need no further iterations
                    if delta_1 - delta_x == 0:
                        break
            
            # Return the calculated deflection
            return delta_x
        
        # Non-P-delta solutions are not iterative
        else:

            # Return the calcuated deflection
            # return delta1 + theta1*x - M1*x**2/(2*EI) + V1*x**3/(6*EI) + w1*x**4/(24*EI) + x**5*(-w1 + w2)/(120*EI*L)
            return delta_1 + theta_1*x + V1*x**3/(6*EI) + w1*x**4/(24*EI) + x**2*(-M1)/(2*EI) + x**5*(-w1 + w2)/(120*EI*L)

    def AxialDeflection(self, x):

        delta_x1 = self.delta_x1
        P1 = self.P1
        p1 = self.p1
        p2 = self.p2
        L = self.Length()
        EA = self.EA

        return delta_x1 - 1/EA*(P1*x + p1*x**2/2 + (p2 - p1)*x**3/(6*L))

    # Returns the maximum shear in the segment
    def max_shear(self):
        
        w1 = self.w1
        w2 = self.w2
        L = self.Length()
        
        # Determine possible locations of maximum shear
        if w1-w2 == 0:
            x1 = 0
        else:
            x1 = w1*L/(w1-w2)
    
        if round(x1, 10) < 0 or round(x1, 10) > round(L, 10):
            x1 = 0
    
        x2 = 0
        x3 = L
    
        # Find the shear at each location of interest
        V1 = self.Shear(x1)
        V2 = self.Shear(x2)
        V3 = self.Shear(x3)
    
        # Return the maximum shear
        return max(V1, V2, V3)

    # Returns the minimum shear in the segment
    def min_shear(self):
        
        w1 = self.w1
        w2 = self.w2
        L = self.Length()
        
        # Determine possible locations of minimum shear
        if w1-w2 == 0:
            x1 = 0
        else:
            x1 = w1*L/(w1-w2)
    
        if round(x1, 10) < 0 or round(x1, 10) > round(L, 10):
            x1 = 0
    
        x2 = 0
        x3 = L
    
        # Find the shear at each location of interest
        V1 = self.Shear(x1)
        V2 = self.Shear(x2)
        V3 = self.Shear(x3)
    
        # Return the minimum shear
        return min(V1, V2, V3)
    
    # Returns the maximum moment in the segment
    def max_moment(self, P_delta=False):
        
        w1 = self.w1
        w2 = self.w2
        V1 = self.V1
        L = self.Length()
        
        # Find the quadratic equation parameters
        a = -(w2-w1)/(2*L)
        b = -w1
        c = -V1
    
        # Determine possible locations of maximum moment
        if a == 0:
            if b != 0:
                x1 = -c/b
            else:
                x1 = 0
            x2 = 0
        elif b**2-4*a*c < 0:
            x1 = 0
            x2 = 0
        else:
            x1 = (-b+(b**2-4*a*c)**0.5)/(2*a)
            x2 = (-b-(b**2-4*a*c)**0.5)/(2*a)
    
        x3 = 0
        x4 = L
    
        if round(x1, 10) < 0 or round(x1, 10) > round(L, 10):
            x1 = 0
    
        if round(x2, 10) < 0 or round(x2, 10) > round(L, 10):
            x2 = 0
    
        # Find the moment at each location of interest
        M1 = self.moment(x1, P_delta)
        M2 = self.moment(x2, P_delta)
        M3 = self.moment(x3, P_delta)
        M4 = self.moment(x4, P_delta)
    
        # Return the maximum moment
        return max(M1, M2, M3, M4)

    # Returns the minimum moment in the segment
    def min_moment(self, P_delta=False):
        
        w1 = self.w1
        w2 = self.w2
        V1 = self.V1
        L = self.Length()
        
        # Find the quadratic equation parameters
        a = -(w2-w1)/(2*L)
        b = -w1
        c = -V1
    
        # Determine possible locations of minimum moment
        if a == 0:
            if b != 0:
                x1 = -c/b
            else:
                x1 = 0
            x2 = 0
        elif b**2-4*a*c < 0:
            x1 = 0
            x2 = 0
        else:
            x1 = (-b+(b**2-4*a*c)**0.5)/(2*a)
            x2 = (-b-(b**2-4*a*c)**0.5)/(2*a)
    
        x3 = 0
        x4 = L
    
        if round(x1, 10) < 0 or round(x1, 10) > round(L, 10):
            x1 = 0
    
        if round(x2, 10) < 0 or round(x2, 10) > round(L, 10):
            x2 = 0
    
        # Find the moment at each location of interest
        M1 = self.moment(x1, P_delta)
        M2 = self.moment(x2, P_delta)
        M3 = self.moment(x3, P_delta)
        M4 = self.moment(x4, P_delta)
    
        # Return the minimum moment
        return min(M1, M2, M3, M4)

    # Returns the maximum axial force in the segment
    def max_axial(self):
        
        p1 = self.p1
        p2 = self.p2
        L = self.Length()
        
        # Determine possible locations of maximum axial force
        if p1-p2 != 0:
            x1 = L*p1/(p1-p2)
        else:
            x1 = 0
    
        if round(x1, 10) < 0 or round(x1, 10) > round(L, 10):
            x1 = 0
    
        x2 = 0
        x3 = L
    
        # Find the axial force at each location of interest
        P1 = self.axial(x1)
        P2 = self.axial(x2)
        P3 = self.axial(x3)
    
        # Return the maximum axial force
        return max(P1, P2, P3)

    # Returns the minimum axial force in the segment
    def min_axial(self):
        
        p1 = self.p1
        p2 = self. p2
        L = self.Length()
        
        # Determine possible locations of minimum axial force
        if p1-p2 != 0:
            x1 = L*p1/(p1-p2)
        else:
            x1 = 0
    
        if round(x1, 10) < 0 or round(x1, 10) > round(L, 10):
            x1 = 0

        x2 = 0
        x3 = L
    
        # Find the axial force at each location of interest
        P1 = self.axial(x1)
        P2 = self.axial(x2)
        P3 = self.axial(x3)
    
        # Return the minimum axial force
        return min(P1, P2, P3)

    def MaxTorsion(self):
        """
        Returns the maximum torsional moment in the segment.
        """
    
        # Return the maximum torsional moment
        # Since the torsional moment is constant on the segment, the maximum torsional moment is T1
        # This can be updated in the future for distributed torsional forces
        return self.T1

    def MinTorsion(self):
        """
        Returns the minimum torsional moment in the segment.
        """

        # Return the minimum torsional moment
        # Since the torsional moment is constant on the segment, the minimum torsional moment is T1
        # This can be updated in the future for distributed torsional forces
        return self.T1




from json import load
import warnings

from IPython.display import Image
from numpy import array, empty, append, cross
from numpy.linalg import norm
import vtk

class Renderer():
    """Used to render finite element models.
    """

    scalar = None

    def __init__(self, model):

        self.model = model

        # Default settings for rendering
        self.annotation_size = 5
        self.deformed_shape = False
        self.deformed_scale = 30
        self.render_nodes = True
        self.render_loads = True
        self.color_map = None
        self.combo_name = 'Combo 1'
        self.case = None
        self.labels = True
        self.scalar_bar = False
        self.scalar_bar_text_size = 24
        self.theme = 'default'

        # Initialize VTK objects
        self.renderer = vtk.vtkRenderer()
        self.window = vtk.vtkRenderWindow()
        self.window.SetWindowName('Pynite - Simple Finite Element Analysis in Python')
        self.window.AddRenderer(self.renderer)

    @property
    def window_width(self):
        return self.window.GetSize()[0]
    
    @window_width.setter
    def window_width(self, width):
        height = self.window.GetSize()[1]
        self.window.SetSize(width, height)
    
    @property
    def window_height(self):
        return self.window.GetSize()[1]
    
    @window_height.setter
    def window_height(self, height):
        width = self.window.GetSize()[0]
        self.window.SetSize(width, height)

    def set_annotation_size(self, size=5):
        self.annotation_size = size
    
    def set_deformed_shape(self, deformed_shape=False):
        self.deformed_shape = deformed_shape
    
    def set_deformed_scale(self, scale=30):
        self.deformed_scale = scale
    
    def set_render_nodes(self, render_nodes=True):
        self.render_nodes = render_nodes

    def set_render_loads(self, render_loads=True):
        self.render_loads = render_loads
    
    def set_color_map(self, color_map=None):
        """Sets the color map for plate contours.

        :param color_map: The color map to use: Valid options are 'Qx', Qy', 'Mx', 'My', 'Mxy', 'Sx', 'Sy' 'Txy'. Qx and Qy are out-of-plane shear forces. Mx, My, and Mxy are local out-of-plane bending moments. 'Sx' and 'Sy' are membrane forces, and 'Txy' is an in-plane shear force. Defaults to None.
        :type color_map: str, optional
        """
        self.color_map = color_map
    
    def set_combo_name(self, combo_name='Combo 1'):
        self.combo_name = combo_name
        self.case = None
    
    def set_case(self, case=None):
        self.case = case
        self.combo_name = None
    
    def set_show_labels(self, show_labels=True):
        self.labels = show_labels
    
    def set_scalar_bar(self, scalar_bar=False):
        self.scalar_bar = scalar_bar
    
    def set_scalar_bar_text_size(self, text_size=24):
        self.scalar_bar_text_size = text_size

    def render_model(self, interact=True, reset_camera=True):
        """
        Renders the model in a window

        Parameters
        ----------
        interact : bool
            Suppresses interacting with the window if set to `False`. This can be used to capture a
            screenshot without pausing the program for the user to interact. Default is `True`.
        reset_camera : bool
            Resets the camera if set to `True`. Default is `True`.
        """
        
        # Get the render window
        window = self.window
        
        # Update the renderer
        self.update(reset_camera)

        # Render the window
        window.Render()

        # Handle user interaction if requested by the user
        if interact:
        
            # Set up an interactor. The interactor style determines how user interactions affect the
            # view. The trackball camera style behaves much like popular commercial CAD programs.
            interactor = vtk.vtkRenderWindowInteractor()
            style = vtk.vtkInteractorStyleTrackballCamera()
            interactor.SetInteractorStyle(style)
            interactor.SetRenderWindow(self.window)
            
            # Start the interactor. Code execution will pause here until the user closes the window.
            interactor.Start()
            
            # Finalize the render window once the user closes out of it. I don't understand everything
            # this does, but I've found screenshots will cause the program to crash if this line is
            # omitted. I have noticed it will shut down the interactor.
            window.Finalize()

        return window

    def screenshot(self, filepath='console', interact=True, reset_camera=True):
        """
        Renders the model in a window. When the window is closed a screenshot is captured.

        Parameters
        ----------
        filepath : string
            Sends a screenshot to the specified filepath. The screenshot will be taken when the
            user closes out of the render window. If `filepath` is set to 'console' the screenshot
            will be returned as an IPython image. If set to 'BytesIO' it will return the image as a
            `BytesIO` object. Default is 'console'.
        interact : bool
            Suppresses interacting with the window if set to `False`. This can be used to capture a
            screenshot without pausing the program for the user to interact. Default is `True`.
        reset_camera : boolju
            Resets the camera if set to `True`. Default is `True`.
        """

        # Render the model in a window and save the window
        window = self.render_model(interact, reset_camera)

        # Screenshot code
        w2if = vtk.vtkWindowToImageFilter()
        w2if.SetInput(window)
        w2if.SetInputBufferTypeToRGB()
        w2if.ReadFrontBufferOff()

        # These next two lines are in the examples and documentation for VTK, but don't seem to do
        # anything. I've left them here in case I find a bug somewhere down the line that needs
        # fixing.
        # w2if.Update()
        # w2if.Modified()

        writer = vtk.vtkPNGWriter()
        writer.SetInputConnection(w2if.GetOutputPort())

        if filepath == 'console' or filepath == 'BytesIO':

            writer.SetWriteToMemory(1)
            writer.Write()
            fig_file = memoryview(writer.GetResult()).tobytes()

            # Now that we're done with the render window, finalize it
            window.Finalize()

            if filepath == 'console':
                return Image(fig_file)
            elif filepath == 'BytesIO':
                from io import BytesIO
                return BytesIO(fig_file)
        else:

            writer.SetFileName(filepath)
            writer.Write()

            # Now that we're done with the render window, finalize it
            window.Finalize()

            return

    def update(self, reset_camera=True):
        """
        Builds or rebuilds the VTK renderer

        Parameters
        ----------
        reset_camera : bool
            Resets the camera if set to `True`. Default is `True`.
        """

        # Input validation
        if self.deformed_shape and self.case != None:
            raise Exception('Deformed shape is only available for load combinations,'
                            ' not load cases.')
        if self.model.load_combos == {} and self.render_loads == True and self.case == None:
            self.render_loads = False
            warnings.warn('Unable to render load combination. No load combinations defined.', UserWarning)
        
        # Check if nodes are to be rendered
        if self.render_nodes == True:

            if self.theme == 'print':
                color = 'black'
            else:
                color = None

            # Create a visual node for each node in the model
            vis_nodes = []
            for node in self.model.nodes.values():
                vis_nodes.append(VisNode(node, self.annotation_size, color))
        
        # Create a visual spring for each spring in the model
        vis_springs = []
        for spring in self.model.springs.values():
            vis_springs.append(VisSpring(spring, self.model.nodes, self.annotation_size))    
    
        # Create a visual member for each member in the model
        vis_members = []
        for member in self.model.members.values():
            vis_members.append(VisMember(member, self.model.nodes, self.annotation_size, self.theme))
        
        # Get the renderer
        renderer = self.renderer

        # Clear out all the old actors from any previous renderings
        for actor in renderer.GetActors():
            renderer.RemoveActor(actor)

        # Add actors for each spring
        for vis_spring in vis_springs:
        
            # Add the actor for the spring
            renderer.AddActor(vis_spring.actor)

            if self.labels == True:
                # Add the actor for the spring label
                renderer.AddActor(vis_spring.lblActor)
            
                # Set the text to follow the camera as the user interacts. This will
                # require a reset of the camera (see below)
                vis_spring.lblActor.SetCamera(renderer.GetActiveCamera())    

        # Add actors for each member
        for vis_member in vis_members:
            
            # Add the actor for the member
            renderer.AddActor(vis_member.actor)

            if self.labels == True:
                
                # Add the actor for the member label
                renderer.AddActor(vis_member.lblActor)
            
                # Set the text to follow the camera as the user interacts. This will
                # require a reset of the camera (see below)
                vis_member.lblActor.SetCamera(renderer.GetActiveCamera())
        
        # Check if nodes are to be rendered
        if self.render_nodes == True:
            
            # Combine the polydata from each node

            # Create an append filter for combining node polydata
            node_polydata = vtk.vtkAppendPolyData()

            for vis_node in vis_nodes:
                
                # Add the node's polydata
                node_polydata.AddInputData(vis_node.polydata.GetOutput())

                if self.labels == True:

                    if self.theme == 'print':
                        vis_node.lblActor.GetProperty().SetColor(0, 0, 0)  # black
                    
                    # Add the actor for the node label
                    renderer.AddActor(vis_node.lblActor)
                
                    # Set the text to follow the camera as the user interacts. This will
                    # require a reset of the camera (see below)
                    vis_node.lblActor.SetCamera(renderer.GetActiveCamera())
            
            # Update the node polydata in the append filter
            node_polydata.Update()
            
            # Create a mapper and actor for the nodes
            node_mapper = vtk.vtkPolyDataMapper()
            node_mapper.SetInputConnection(node_polydata.GetOutputPort())
            node_actor = vtk.vtkActor()
            node_actor.SetMapper(node_mapper)

            # Adjust the color of the nodes here.
            if self.theme == 'print':
                node_actor.GetProperty().SetColor(0, 0, 0)  # Black
            
            # Add the node actor to the renderer
            renderer.AddActor(node_actor)

        # Render the deformed shape if requested
        if self.deformed_shape == True:
            _DeformedShape(self.model, renderer, self.deformed_scale, self.annotation_size, self.combo_name, self.render_nodes, self.theme)

        # Render the loads if requested
        if (self.combo_name != None or self.case != None) and self.render_loads != False:
            _RenderLoads(self.model, renderer, self.annotation_size, self.combo_name, self.case, self.theme)
        
        # Render the plates and quads, if present
        if self.model.quads or self.model.plates:
            _RenderContours(self.model, renderer, self.deformed_shape, self.deformed_scale,
                            self.color_map, self.scalar_bar, self.scalar_bar_text_size,
                            self.combo_name, self.theme)

        # Set the window's background color
        if self.theme == 'default':
            renderer.SetBackground(0, 0, 128)  # Blue
        elif self.theme == 'print':
            renderer.SetBackground(255, 255, 255)  # White
        
        # Reset the camera
        if reset_camera: renderer.ResetCamera()

#%%
# Converts a node object into a node for the viewer
class VisNode():

    # Constructor
    def __init__(self, node, annotation_size=5, color=None):
      
        # Create an append filter to append all the sources related to the node into a single 'PolyData' object
        self.polydata = vtk.vtkAppendPolyData()
      
        # Get the node's position
        X = node.X # Global X coordinate
        Y = node.Y # Global Y coordinate
        Z = node.Z # Global Z coordinate
      
        # Generate a sphere source for the node
        sphere = vtk.vtkSphereSource()
        sphere.SetCenter(X, Y, Z)
        sphere.SetRadius(0.6*annotation_size)
        sphere.Update()
        self.polydata.AddInputData(sphere.GetOutput())
      
        # Create the text for the node label
        label = vtk.vtkVectorText()
        label.SetText(node.name)
        
        # Set up a mapper for the node label
        lblMapper = vtk.vtkPolyDataMapper()
        lblMapper.SetInputConnection(label.GetOutputPort())
      
        # Set up an actor for the node label
        self.lblActor = vtk.vtkFollower()
        self.lblActor.SetMapper(lblMapper)
        self.lblActor.SetScale(annotation_size, annotation_size, annotation_size)
        self.lblActor.SetPosition(X + 0.6*annotation_size, Y + 0.6*annotation_size, Z)
      
        # Generate any supports that occur at the node
        # Check for a fixed suppport
        if node.support_DX == True and node.support_DY == True and node.support_DZ == True \
        and node.support_RX == True and node.support_RY == True and node.support_RZ == True:
      
            # Create the fixed support
            support = vtk.vtkCubeSource()
            support.SetCenter(node.X, node.Y, node.Z)
            support.SetXLength(annotation_size*1.2)
            support.SetYLength(annotation_size*1.2)
            support.SetZLength(annotation_size*1.2)
        
            # Copy and append the support data to the append filter
            support.Update()
            self.polydata.AddInputData(support.GetOutput())
        
        # Check for a pinned support
        elif node.support_DX == True and node.support_DY == True and node.support_DZ == True \
        and node.support_RX == False and node.support_RY == False and node.support_RZ == False:
          
            # Create the pinned support
            support = vtk.vtkConeSource()
            support.SetCenter(node.X, node.Y-0.6*annotation_size, node.Z)
            support.SetDirection((0, 1, 0))
            support.SetHeight(annotation_size*1.2)
            support.SetRadius(annotation_size*1.2)
        
            # Copy and append the support data to the append filter
            support.Update()
            self.polydata.AddInputData(support.GetOutput())
        
        # Other support conditions
        else:
      
            # Restrained against X translation
            if node.support_DX == True:
                  
                # Create the support
                support1 = vtk.vtkLineSource()  # The line showing the support direction
                support1.SetPoint1(node.X-annotation_size, node.Y, node.Z)
                support1.SetPoint2(node.X+annotation_size, node.Y, node.Z)
          
                # Copy and append the support data to the append filter
                support1.Update()
                self.polydata.AddInputData(support1.GetOutput())
          
                support2 = vtk.vtkConeSource()
                support2.SetCenter(node.X-annotation_size, node.Y, node.Z)
                support2.SetDirection((1, 0, 0))
                support2.SetHeight(annotation_size*0.6)
                support2.SetRadius(annotation_size*0.3)
          
                # Copy and append the support data to the append filter
                support2.Update()
                self.polydata.AddInputData(support2.GetOutput())
          
                support3 = vtk.vtkConeSource()
                support3.SetCenter(node.X+annotation_size, node.Y, node.Z)
                support3.SetDirection((-1, 0, 0))
                support3.SetHeight(annotation_size*0.6)
                support3.SetRadius(annotation_size*0.3)
          
                # Copy and append the support data to the append filter
                support3.Update()
                self.polydata.AddInputData(support3.GetOutput())
          
            # Restrained against Y translation
            if node.support_DY == True:
              
                # Create the support
                support1 = vtk.vtkLineSource()  # The line showing the support direction
                support1.SetPoint1(node.X, node.Y-annotation_size, node.Z)
                support1.SetPoint2(node.X, node.Y+annotation_size, node.Z)
          
                # Copy and append the support data to the append filter
                support1.Update()
                self.polydata.AddInputData(support1.GetOutput())
          
                support2 = vtk.vtkConeSource()
                support2.SetCenter(node.X, node.Y-annotation_size, node.Z)
                support2.SetDirection((0, 1, 0))
                support2.SetHeight(annotation_size*0.6)
                support2.SetRadius(annotation_size*0.3)
          
                # Copy and append the support data to the append filter
                support2.Update()
                self.polydata.AddInputData(support2.GetOutput())
          
                support3 = vtk.vtkConeSource()
                support3.SetCenter(node.X, node.Y+annotation_size, node.Z)
                support3.SetDirection((0, -1, 0))
                support3.SetHeight(annotation_size*0.6)
                support3.SetRadius(annotation_size*0.3)
          
                # Copy and append the support data to the append filter
                support3.Update()
                self.polydata.AddInputData(support3.GetOutput())
          
            # Restrained against Z translation
            if node.support_DZ == True:
            
                # Create the support
                support1 = vtk.vtkLineSource()  # The line showing the support direction
                support1.SetPoint1(node.X, node.Y, node.Z-annotation_size)
                support1.SetPoint2(node.X, node.Y, node.Z+annotation_size)
          
                # Copy and append the support data to the append filter
                support1.Update()
                self.polydata.AddInputData(support1.GetOutput())
          
                support2 = vtk.vtkConeSource()
                support2.SetCenter(node.X, node.Y, node.Z-annotation_size)
                support2.SetDirection((0, 0, 1))
                support2.SetHeight(annotation_size*0.6)
                support2.SetRadius(annotation_size*0.3)
          
                # Copy and append the support data to the append filter
                support2.Update()
                self.polydata.AddInputData(support2.GetOutput())
          
                support3 = vtk.vtkConeSource()
                support3.SetCenter(node.X, node.Y, node.Z+annotation_size)
                support3.SetDirection((0, 0, -1))
                support3.SetHeight(annotation_size*0.6)
                support3.SetRadius(annotation_size*0.3)
          
                # Copy and append the support data to the append filter
                support3.Update()
                self.polydata.AddInputData(support3.GetOutput())
      
            # Restrained against rotation about the X-axis
            if node.support_RX == True:
            
                # Create the support
                support1 = vtk.vtkLineSource()  # The line showing the support direction
                support1.SetPoint1(node.X-1.6*annotation_size, node.Y, node.Z)
                support1.SetPoint2(node.X+1.6*annotation_size, node.Y, node.Z)
          
                # Copy and append the support data to the append filter
                support1.Update()
                self.polydata.AddInputData(support1.GetOutput())
          
                support2 = vtk.vtkCubeSource()
                support2.SetCenter(node.X-1.9*annotation_size, node.Y, node.Z)
                support2.SetXLength(annotation_size*0.6)
                support2.SetYLength(annotation_size*0.6)
                support2.SetZLength(annotation_size*0.6)
          
                # Copy and append the support data to the append filter
                support2.Update()
                self.polydata.AddInputData(support2.GetOutput())
          
                support3 = vtk.vtkCubeSource()
                support3.SetCenter(node.X+1.9*annotation_size, node.Y, node.Z)
                support3.SetXLength(annotation_size*0.6)
                support3.SetYLength(annotation_size*0.6)
                support3.SetZLength(annotation_size*0.6)
          
                # Copy and append the support data to the append filter
                support3.Update()
                self.polydata.AddInputData(support3.GetOutput())
          
            # Restrained against rotation about the Y-axis
            if node.support_RY == True:
            
                # Create the support
                support1 = vtk.vtkLineSource()  # The line showing the support direction
                support1.SetPoint1(node.X, node.Y-1.6*annotation_size, node.Z)
                support1.SetPoint2(node.X, node.Y+1.6*annotation_size, node.Z)
          
                # Copy and append the support data to the append filter
                support1.Update()
                self.polydata.AddInputData(support1.GetOutput())
          
                support2 = vtk.vtkCubeSource()
                support2.SetCenter(node.X, node.Y-1.9*annotation_size, node.Z)
                support2.SetXLength(annotation_size*0.6)
                support2.SetYLength(annotation_size*0.6)
                support2.SetZLength(annotation_size*0.6)
          
                # Copy and append the support data to the append filter
                support2.Update()
                self.polydata.AddInputData(support2.GetOutput())
          
                support3 = vtk.vtkCubeSource()
                support3.SetCenter(node.X, node.Y+1.9*annotation_size, node.Z)
                support3.SetXLength(annotation_size*0.6)
                support3.SetYLength(annotation_size*0.6)
                support3.SetZLength(annotation_size*0.6)
          
                # Copy and append the support data to the append filter
                support3.Update()
                self.polydata.AddInputData(support3.GetOutput())
          
            # Restrained against rotation about the Z-axis
            if node.support_RZ == True:
            
                # Create the support
                support1 = vtk.vtkLineSource()  # The line showing the support direction
                support1.SetPoint1(node.X, node.Y, node.Z-1.6*annotation_size)
                support1.SetPoint2(node.X, node.Y, node.Z+1.6*annotation_size)
          
                # Copy and append the support data to the append filter
                support1.Update()
                self.polydata.AddInputData(support1.GetOutput())
          
                support2 = vtk.vtkCubeSource()
                support2.SetCenter(node.X, node.Y, node.Z-1.9*annotation_size)
                support2.SetXLength(annotation_size*0.6)
                support2.SetYLength(annotation_size*0.6)
                support2.SetZLength(annotation_size*0.6)
          
                # Copy and append the support data to the append filter
                support2.Update()
                self.polydata.AddInputData(support2.GetOutput())
          
                support3 = vtk.vtkCubeSource()
                support3.SetCenter(node.X, node.Y, node.Z+1.9*annotation_size)
                support3.SetXLength(annotation_size*0.6)
                support3.SetYLength(annotation_size*0.6)
                support3.SetZLength(annotation_size*0.6)
          
                # Copy and append the support data to the append filter
                support3.Update()
                self.polydata.AddInputData(support3.GetOutput())
        
        # Update the append filter
        self.polydata.Update()
      
        # Create a mapper and actor
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(self.polydata.GetOutputPort())
        self.actor = vtk.vtkActor()
        
        # Set the mapper for the node's actor
        self.actor.SetMapper(mapper)

        # Color will be added to the node actors outside of this class once they are all combined into one `vtkAppendFilter` object.

        # TODO: Delete legacy code below once it's been proven to be unnecessary over time.

        # Add color to the node and label actors if specified
        # if color == 'red':
        #     self.actor.GetProperty().SetColor(255, 0, 0)  # Red
        #     self.lblActor.GetProperty().SetColor(255, 0, 0)  # Red
        # elif color == 'yellow':
        #     self.actor.GetProperty().SetColor(255, 255, 0)  # Yellow
        #     self.lblActor.GetProperty().SetColor(255, 255, 0)  # Yellow
        # elif color == 'black':
        #     self.actor.GetProperty().SetColor(0, 0, 0)  # Black
        #     self.lblActor.GetProperty().SetColor(0, 0, 0)  # Black

class VisSpring():
    
    def __init__(self, spring, nodes, annotation_size=5, color=None):

        # Generate a line source for the spring
        line = vtk.vtkLineSource()

        # Step through each node in the model and find the position of the
        # i-node and j-node
        for node in nodes.values():

            # Check to see if the current node is the i-node
            if node.name == spring.i_node.name:
                Xi = node.X
                Yi = node.Y
                Zi = node.Z
                line.SetPoint1(Xi, Yi, Zi)

            # Check to see if the current node is the j-node
            elif node.name == spring.j_node.name:
                Xj = node.X
                Yj = node.Y
                Zj = node.Z
                line.SetPoint2(Xj, Yj, Zj)
    
        # Set up a mapper for the spring
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(line.GetOutputPort())

        # Set up an actor and a mapper for the spring
        self.actor = vtk.vtkActor()
        self.actor.SetMapper(mapper)

        # Create the text for the spring label
        label = vtk.vtkVectorText()
        label.SetText(spring.name)

        # Set up a mapper for the spring label
        lblMapper = vtk.vtkPolyDataMapper()
        lblMapper.SetInputConnection(label.GetOutputPort())

        # Set up an actor for the spring label
        self.lblActor = vtk.vtkFollower()
        self.lblActor.SetMapper(lblMapper)
        self.lblActor.SetScale(annotation_size, annotation_size, annotation_size)
        self.lblActor.SetPosition((Xi+Xj)/2, (Yi+Yj)/2, (Zi+Zj)/2)

        # Add some color
        if color is None:
            self.actor.GetProperty().SetColor(255, 0, 255)     # Magenta
            self.lblActor.GetProperty().SetColor(255, 0, 255)
        elif color == 'black':
            self.actor.GetProperty().SetColor(0, 0, 0)         # Black
            self.lblActor.GetProperty().SetColor(0, 0, 0)
      
# Converts a member object into a member for the viewer
class VisMember():

    # Constructor
    def __init__(self, member, nodes, annotation_size=5, theme='default'):
    
        # Generate a line for the member
        line = vtk.vtkLineSource()
      
        # Step through each node in the model and find the position of the i-node and j-node
        for node in nodes.values():
        
            # Check to see if the current node is the i-node
            if node.name == member.i_node.name:
                Xi = node.X
                Yi = node.Y
                Zi = node.Z
                line.SetPoint1(Xi, Yi, Zi)
        
            # Check to see if the current node is the j-node
            elif node.name == member.j_node.name:
                Xj = node.X
                Yj = node.Y
                Zj = node.Z
                line.SetPoint2(Xj, Yj, Zj)
        
        # Set up a mapper for the member
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(line.GetOutputPort())
      
        # Set up an actor for the member
        self.actor = vtk.vtkActor()
        self.actor.SetMapper(mapper)
      
        # Create the text for the member label
        label = vtk.vtkVectorText()
        label.SetText(member.name)
      
        # Set up a mapper for the member label
        lblMapper = vtk.vtkPolyDataMapper()
        lblMapper.SetInputConnection(label.GetOutputPort())
      
        # Set up an actor for the member label
        self.lblActor = vtk.vtkFollower()
        self.lblActor.SetMapper(lblMapper)
        self.lblActor.SetScale(annotation_size, annotation_size, annotation_size)
        self.lblActor.SetPosition((Xi+Xj)/2, (Yi+Yj)/2, (Zi+Zj)/2)

        # Adjust the color of the member if the theme is 'print'
        if theme == 'print':
            self.actor.GetProperty().SetColor(0/255, 0/255, 0/255)  # Black
            self.lblActor.GetProperty().SetColor(0/255, 0/255, 0/255) # Black

# Converts a node object into a node in its deformed position for the viewer
class VisDeformedNode():
    
    def __init__(self, node, scale_factor, annotation_size=5, combo_name='Combo 1'):
          
        # Calculate the node's deformed position
        newX = node.X + scale_factor*(node.DX[combo_name])
        newY = node.Y + scale_factor*(node.DY[combo_name])
        newZ = node.Z + scale_factor*(node.DZ[combo_name])
            
        # Generate a sphere source for the node in its deformed position
        self.source = vtk.vtkSphereSource()
        self.source.SetCenter(newX, newY, newZ)
        self.source.SetRadius(0.6*annotation_size)
        self.source.Update()

class VisDeformedMember():
    
    def __init__(self, member, nodes, scale_factor, combo_name='Combo 1'):
        
        # Determine if this member is active for each load combination
        self.active = member.active
      
        L = member.L() # Member length
        T = member.T() # Member local transformation matrix
      
        cos_x = array([T[0,0:3]]) # Direction cosines of local x-axis
        cos_y = array([T[1,0:3]]) # Direction cosines of local y-axis
        cos_z = array([T[2,0:3]]) # Direction cosines of local z-axis
      
        # Find the initial position of the local i-node
        # Step through each node
        for node in nodes.values():
          
            # Check to see if the current node is the i-node
            if node.name == member.i_node.name:
                Xi = node.X
                Yi = node.Y
                Zi = node.Z
      
        # Calculate the local y-axis displacements at 20 points along the member's
        # length
        DY_plot = empty((0, 3))
        for i in range(20):
                
            # Calculate the local y-direction displacement
            dy_tot = member.deflection('dy', L/19*i, combo_name)
        
            # Calculate the scaled displacement in global coordinates
            DY_plot = append(DY_plot, dy_tot*cos_y*scale_factor, axis=0)
      
        # Calculate the local z-axis displacements at 20 points along the member's
        # length
        DZ_plot = empty((0, 3)) 
        for i in range(20):
                
            # Calculate the local z-direction displacement
            dz_tot = member.deflection('dz', L/19*i, combo_name)
        
            # Calculate the scaled displacement in global coordinates
            DZ_plot = append(DZ_plot, dz_tot*cos_z*scale_factor, axis=0)
      
        # Calculate the local x-axis displacements at 20 points along the member's
        # length
        DX_plot = empty((0, 3)) 
        for i in range(20):
                
            # Displacements in local coordinates
            dx_tot = [[Xi, Yi, Zi]] + (L/19*i + member.deflection('dx', L/19*i, combo_name)*scale_factor)*cos_x
                  
            # Magnified displacements in global coordinates
            DX_plot = append(DX_plot, dx_tot, axis=0)
        
        # Sum the component displacements to obtain overall displacement
        D_plot = DY_plot + DZ_plot + DX_plot
      
        # Generate vtk points
        points = vtk.vtkPoints()
        points.SetNumberOfPoints(len(D_plot))
      
        for i in range(len(D_plot)):
            points.SetPoint(i, D_plot[i, 0], D_plot[i, 1], D_plot[i, 2])
      
        # Generate vtk lines
        lines = vtk.vtkCellArray()
        lines.InsertNextCell(len(D_plot))
      
        for i in range(len(D_plot)):
            lines.InsertCellPoint(i)
      
        # Create a polyline source from the defined points and lines
        self.source = vtk.vtkPolyData()
        self.source.SetPoints(points)
        self.source.SetLines(lines)   
                                  
class VisDeformedSpring():
    
    def __init__(self, spring, nodes, scale_factor, combo_name='Combo 1'):

        # Determine if this spring is active for each load combination
        self.active = spring.active
        
        # Generate a line source for the spring
        self.source = vtk.vtkLineSource()
        
        # Find the deformed position of the local i-node
        # Step through each node
        for node in nodes.values():
      
            # Check to see if the current node is the i-node
            if node.name == spring.i_node.name:
                Xi = node.X + node.DX[combo_name]*scale_factor
                Yi = node.Y + node.DY[combo_name]*scale_factor
                Zi = node.Z + node.DZ[combo_name]*scale_factor
                self.source.SetPoint1(Xi, Yi, Zi)
        
            # Check to see if the current node is the i-node
            if node.name == spring.j_node.name:
                Xj = node.X + node.DX[combo_name]*scale_factor
                Yj = node.Y + node.DY[combo_name]*scale_factor
                Zj = node.Z + node.DZ[combo_name]*scale_factor
                self.source.SetPoint2(Xj, Yj, Zj)
        
        self.source.Update()
    
class VisPtLoad():
    '''
    Creates a point load for the viewer
    '''
    
    def __init__(self, position, direction, length, label_text=None, annotation_size=5, color=None):
        '''
        Constructor.
      
        Parameters
        ----------
        position : tuple
          A tuple of X, Y and Z coordinates for the point of the load arrow: (X, Y, Z).
        direction : tuple
          A tuple indicating the direction vector for the load arrow: (i, j, k).
        length : number
          The length of the load arrow.
        tip_length : number
          The height of the arrow head.
        label_text : string
          Text that will show up at the tail of the arrow. If set to 'None' no text will be displayed.
        '''
      
        # Create a unit vector in the direction of the 'direction' vector
        unitVector = direction/norm(direction)
      
        # Create a 'vtkAppendPolyData' filter to append the tip and shaft together into a single dataset
        self.polydata = vtk.vtkAppendPolyData()
      
        # Determine if the load is positive or negative
        if length == 0:
            sign = 1
        else:
            sign = abs(length)/length
      
        # Generate the tip of the load arrow
        tip_length = abs(length)/4
        radius = abs(length)/16
        tip = vtk.vtkConeSource()
        tip.SetCenter(position[0] - tip_length*sign*0.5*unitVector[0], \
                      position[1] - tip_length*sign*0.5*unitVector[1], \
                      position[2] - tip_length*sign*0.5*unitVector[2])
        tip.SetDirection([direction[0]*sign, direction[1]*sign, direction[2]*sign])
        tip.SetHeight(tip_length)
        tip.SetRadius(radius)
        tip.Update()
      
        # Add the arrow tip to the append filter
        self.polydata.AddInputData(tip.GetOutput())
        
        # Create the shaft
        shaft = vtk.vtkLineSource()
        shaft.SetPoint1(position)
        shaft.SetPoint2((position[0]-length*unitVector[0], position[1]-length*unitVector[1], position[2]-length*unitVector[2]))
        shaft.Update()
      
        # Copy and append the shaft data to the append filter
        self.polydata.AddInputData(shaft.GetOutput())
        self.polydata.Update()
      
        # Create a mapper and actor
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(self.polydata.GetOutputPort())
        self.actor = vtk.vtkActor()
        if color is None: self.actor.GetProperty().SetColor(0, 255, 0) # Green
        elif color == 'black': self.actor.GetProperty().SetColor(0, 0, 0)  # Black
        self.actor.SetMapper(mapper)
      
        # Create the label if needed
        if label_text != None:
      
            # Create the label and set its text
            self.label = vtk.vtkVectorText()
            self.label.SetText(label_text)
        
            # Set up a mapper for the label
            lblMapper = vtk.vtkPolyDataMapper()
            lblMapper.SetInputConnection(self.label.GetOutputPort())
        
            # Set up an actor for the label
            self.lblActor = vtk.vtkFollower()
            self.lblActor.SetMapper(lblMapper)
            self.lblActor.SetScale(annotation_size, annotation_size, annotation_size)
            self.lblActor.SetPosition(position[0] - (length - 0.6*annotation_size)*unitVector[0], \
                                      position[1] - (length - 0.6*annotation_size)*unitVector[1], \
                                      position[2] - (length - 0.6*annotation_size)*unitVector[2])
            if color is None: self.lblActor.GetProperty().SetColor(0, 255, 0)  # Green
            elif color == 'black': self.lblActor.GetProperty().SetColor(0, 0, 0)  # Black
      
class VisDistLoad():
    '''
    Creates a distributed load for the viewer
    '''
    
    def __init__(self, position1, position2, direction, length1, length2, label_text1, label_text2, annotation_size=5, color=None):
        '''
        Constructor.
        '''
      
        # Calculate the length of the distributed load
        loadLength = ((position2[0]-position1[0])**2 + (position2[1]-position1[1])**2 + (position2[2]-position1[2])**2)**0.5
      
        # Find the direction cosines for the line the load acts on
        lineDirCos = [(position2[0]-position1[0])/loadLength, (position2[1]-position1[1])/loadLength, (position2[2]-position1[2])/loadLength]
      
        # Find the direction cosines for the direction the load acts in
        dirDirCos = direction/norm(direction)
      
        # Create point loads at intervals roughly equal to 75% of the load's largest length (magnitude)
        # Add text labels to the first and last load arrow
        if loadLength > 0:
            num_steps = int(round(0.75*loadLength/max(abs(length1), abs(length2)), 0))
        else:
            num_steps = 0

        num_steps = max(num_steps, 1)
        step = loadLength/num_steps
        ptLoads = []
      
        for i in range(num_steps + 1):
      
            # Calculate the position (X, Y, Z) of this load arrow's point
            position = (position1[0] + i*step*lineDirCos[0], position1[1] + i*step*lineDirCos[1], position1[2] + i*step*lineDirCos[2])
        
            # Determine the length of this load arrow
            length = length1 + (length2 - length1)/loadLength*i*step
        
            # Determine the label's text
            if i == 0:
                label_text = label_text1
            elif i == num_steps:
                label_text = label_text2
        
            # Create the load arrow
            ptLoads.append(VisPtLoad(position, direction, length, label_text, annotation_size=annotation_size))
          
        # Draw a line between the first and last load arrow's tails
        tail_line = vtk.vtkLineSource()
        tail_line.SetPoint1((position1[0] - length1*dirDirCos[0], position1[1] - length1*dirDirCos[1], position1[2] - length1*dirDirCos[2]))
        tail_line.SetPoint2((position2[0] - length2*dirDirCos[0], position2[1] - length2*dirDirCos[1], position2[2] - length2*dirDirCos[2]))
      
        # Combine all the geometry into one 'vtkPolyData' object
        self.polydata = vtk.vtkAppendPolyData()
        for arrow in ptLoads:
            arrow.polydata.Update()
            self.polydata.AddInputData(arrow.polydata.GetOutput())
        
        tail_line.Update()
        self.polydata.AddInputData(tail_line.GetOutput())
        self.polydata.Update()
      
        # Create a mapper and actor for the geometry
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(self.polydata.GetOutputPort())
        self.actor = vtk.vtkActor()
        if color is None: self.actor.GetProperty().SetColor(0, 255, 0)  # Green
        elif color == 'black': self.actor.GetProperty().SetColor(0, 0, 0)  # Black
        self.actor.SetMapper(mapper)
      
        # Get the actors for the labels
        self.lblActors = [ptLoads[0].lblActor, ptLoads[len(ptLoads) - 1].lblActor]

class VisMoment():
    '''
    Creates a concentrated moment for the viewer
    '''
    
    def __init__(self, center, direction, radius, label_text=None, annotation_size=5, color=None):
        '''
        Constructor.
      
        Parameters
        ----------
        center : tuple
          A tuple of X, Y and Z coordinates for center of the moment: (X, Y, Z).
        direction : tuple
          A tuple indicating the direction vector for the moment: (i, j, k).
        radius : number
          The radius of the moment.
        tip_length : number
          The height of the arrow head.
        label_text : string
          Text that will show up at the tail of the moment. If set to 'None' no text will be displayed.
        '''
      
        # Create an append filter to store load polydata in
        self.polydata = vtk.vtkAppendPolyData()
        
        # Find a vector perpendicular to the directional unit vector
        v1 = direction/norm(direction)  # v1 = The directional unit vector for the moment
        v2 = _PerpVector(v1)             # v2 = A unit vector perpendicular to v1
        v3 = cross(v1, v2)
        v3 = v3/norm(v3)                # v3 = A unit vector perpendicular to v1 and v2
      
        # Generate an arc for the moment
        Xc, Yc, Zc = center
        arc = vtk.vtkArcSource()
        arc.SetCenter(Xc, Yc, Zc)
        arc.SetPoint1(Xc + v2[0]*radius, Yc + v2[1]*radius, Zc + v2[2]*radius)
        arc.SetPoint2(Xc + v3[0]*radius, Yc + v3[1]*radius, Zc + v3[2]*radius)
        arc.SetNegative(True)
        arc.SetResolution(20)
        arc.Update()
        self.polydata.AddInputData(arc.GetOutput())
      
        # Generate the arrow tip at the end of the arc
        tip_length = radius/2
        cone_radius = radius/8
        tip = vtk.vtkConeSource()
        tip.SetCenter(arc.GetPoint1()[0], arc.GetPoint1()[1], arc.GetPoint1()[2])
        tip.SetDirection(cross(v1, v2))
        tip.SetHeight(tip_length)
        tip.SetRadius(cone_radius)
        tip.Update()
        self.polydata.AddInputData(tip.GetOutput())
      
        # Update the polydata one last time now that we're done appending items to it
        self.polydata.Update()
      
        # Create the text label
        label = vtk.vtkVectorText()
        label.SetText(label_text)
        lblMapper = vtk.vtkPolyDataMapper()
        lblMapper.SetInputConnection(label.GetOutputPort())
        self.lblActor = vtk.vtkFollower()
        self.lblActor.SetMapper(lblMapper)
        self.lblActor.SetScale(annotation_size, annotation_size, annotation_size)
        self.lblActor.SetPosition(Xc + v3[0]*(radius + 0.25*annotation_size), \
                                  Yc + v3[1]*(radius + 0.25*annotation_size), \
                                  Zc + v3[2]*(radius + 0.25*annotation_size))
        if color is None: self.lblActor.GetProperty().SetColor(0, 255, 0)  # Green
        elif color == 'black': self.lblActor.GetProperty().SetColor(0, 0, 0)  # Black

class VisAreaLoad():
    '''
    Creates an area load for the viewer
    '''
    
    def __init__(self, position0, position1, position2, position3, direction, length, label_text, annotation_size=5, theme='default'):
        '''
        Constructor
        '''
      
        # Create a point load for each corner of the area load
        ptLoads = []
        ptLoads.append(VisPtLoad(position0, direction, length, label_text, annotation_size=annotation_size))
        ptLoads.append(VisPtLoad(position1, direction, length, label_text, annotation_size=annotation_size))
        ptLoads.append(VisPtLoad(position2, direction, length, label_text, annotation_size=annotation_size))
        ptLoads.append(VisPtLoad(position3, direction, length, label_text, annotation_size=annotation_size))
      
        # Find the direction cosines for the direction the load acts in
        dirDirCos = direction/norm(direction)
      
        # Find the positions of the tails of all the arrows at the corners of the area load. This is
        # where we will place the polygon.
        self.p0 = position0 - dirDirCos*length
        self.p1 = position1 - dirDirCos*length
        self.p2 = position2 - dirDirCos*length
        self.p3 = position3 - dirDirCos*length
      
        # Combine all geometry into one 'vtkPolyData' object
        self.polydata = vtk.vtkAppendPolyData()
        for arrow in ptLoads:
            self.polydata.AddInputData(arrow.polydata.GetOutput())
        self.polydata.Update()
      
        # Add a label
        self.label_actor = ptLoads[0].lblActor

        # Add color to the area load label
        if theme == 'print':
            self.label_actor.GetProperty().SetColor(255/255, 0/255, 0/255)  # red
        elif theme == 'default':
            self.label_actor.GetProperty().SetColor(0/255, 255/255, 0/255)  # green

def _PerpVector(v):
    '''
    Returns a unit vector perpendicular to v=[i, j, k]
    '''
    
    i = v[0]
    j = v[1]
    k = v[2]
    
    # Find a vector in a direction perpendicular to <i, j, k>
    if i == 0:
        i2 = 1
        j2 = 0
        k2 = 0
    elif j == 0:
        i2 = 0
        j2 = 1
        k2 = 0
    elif k == 0:
        i2 = 0
        j2 = 0
        k2 = 1
    else:
        i2 = 1
        j2 = 1
        k2 = -(i*i2+j*j2)/k
    
    # Return the unit vector
    return [i2, j2, k2]/norm([i2, j2, k2])

def _PrepContour(model, stress_type='Mx', combo_name='Combo 1'):

    if stress_type != None:
    
        # Erase any previous contours
        for node in model.nodes.values():
            node.contour = []
    
        # Step through each element in the model
        for element in list(model.quads.values()) + list(model.plates.values()):
            
            # Rectangular elements and quadrilateral elements have different local coordinate systems.
            # Rectangles are based on a traditional (x, y) system, while quadrilaterals are based on a
            # 'natural' (r, s) coordinate system. To reduce duplication of code for both these elements
            # we'll define the edges of the plate here for either element using the (r, s) terminology.
            if element.type == 'Rect':
                r_left = 0
                r_right = element.width()
                s_bot = 0
                s_top = element.height()
            else:
                r_left = -1
                r_right = 1
                s_bot = -1
                s_top = 1
      
            # Determine which stress result has been requested by the user
            if stress_type == 'dz':
                # Internally Pynite defines the nodes for a rectangular element in the order (i, j, m, n),
                # while it defines the nodes for a quadrilateral element in the order (m, n, i, j)
                if element.type == 'Rect':
                    i, j, m, n = element.d(combo_name)[[2, 8, 14, 20], :]
                else:
                    i, j, m, n = element.d(combo_name)[[14, 20, 2, 8], :]
                element.i_node.contour.append(i)
                element.j_node.contour.append(j)
                element.m_node.contour.append(m)
                element.n_node.contour.append(n)
            elif stress_type == 'Mx':
                element.i_node.contour.append(element.moment(r_left, s_bot, combo_name)[0])
                element.j_node.contour.append(element.moment(r_right, s_bot, combo_name)[0])
                element.m_node.contour.append(element.moment(r_right, s_top, combo_name)[0])
                element.n_node.contour.append(element.moment(r_left, s_top, combo_name)[0])
            elif stress_type == 'My':
                element.i_node.contour.append(element.moment(r_left, s_bot, combo_name)[1])
                element.j_node.contour.append(element.moment(r_right, s_bot, combo_name)[1])
                element.m_node.contour.append(element.moment(r_right, s_top, combo_name)[1])
                element.n_node.contour.append(element.moment(r_left, s_top, combo_name)[1])
            elif stress_type == 'Mxy':
                element.i_node.contour.append(element.moment(r_left, s_bot, combo_name)[2])
                element.j_node.contour.append(element.moment(r_right, s_bot, combo_name)[2])
                element.m_node.contour.append(element.moment(r_right, s_top, combo_name)[2])
                element.n_node.contour.append(element.moment(r_left, s_top, combo_name)[2])
            elif stress_type == 'Qx':
                element.i_node.contour.append(element.shear(r_left, s_bot, combo_name)[0])
                element.j_node.contour.append(element.shear(r_right, s_bot, combo_name)[0])
                element.m_node.contour.append(element.shear(r_right, s_top, combo_name)[0])
                element.n_node.contour.append(element.shear(r_left, s_top, combo_name)[0])
            elif stress_type == 'Qy':
                element.i_node.contour.append(element.shear(r_left, s_bot, combo_name)[1])
                element.j_node.contour.append(element.shear(r_right, s_bot, combo_name)[1])
                element.m_node.contour.append(element.shear(r_right, s_top, combo_name)[1])
                element.n_node.contour.append(element.shear(r_left, s_top, combo_name)[1])
            elif stress_type == 'Sx':
                element.i_node.contour.append(element.membrane(r_left, s_bot, combo_name)[0])
                element.j_node.contour.append(element.membrane(r_right, s_bot, combo_name)[0])
                element.m_node.contour.append(element.membrane(r_right, s_top, combo_name)[0])
                element.n_node.contour.append(element.membrane(r_left, s_top, combo_name)[0])
            elif stress_type == 'Sy':
                element.i_node.contour.append(element.membrane(r_left, s_bot, combo_name)[1])
                element.j_node.contour.append(element.membrane(r_right, s_bot, combo_name)[1])
                element.m_node.contour.append(element.membrane(r_right, s_top, combo_name)[1])
                element.n_node.contour.append(element.membrane(r_left, s_top, combo_name)[1])
            elif stress_type == 'Txy':
                element.i_node.contour.append(element.membrane(r_left, s_bot, combo_name)[2])
                element.j_node.contour.append(element.membrane(r_right, s_bot, combo_name)[2])
                element.m_node.contour.append(element.membrane(r_right, s_top, combo_name)[2])
                element.n_node.contour.append(element.membrane(r_left, s_top, combo_name)[2])                 
      
        # Average the values at each node to obtain a smoothed contour
        for node in model.nodes.values():
            # Prevent divide by zero errors for nodes with no contour values
            if node.contour != []:
                node.contour = sum(node.contour)/len(node.contour)

def _DeformedShape(model, vtk_renderer, scale_factor, annotation_size, combo_name, render_nodes=True, theme='default'):
    '''
    Renders the deformed shape of a model.
    
    Parameters
    ----------
    model : FEModel3D
        Finite element model to be rendered.
    renderer : vtk.vtkRenderer
        The VTK renderer object that will render the model.
    scale_factor : number
        The scale factor to apply to the model deformations.
    annotation_size : number
        Controls the height of text displayed with the model. The units used for `annotation_size` are
        the same as those used for lengths in the model. Sizes of other objects (such as nodes) are
        related to this value.
    combo_name : string
        The load case used for rendering the deflected shape.
    
    Returns
    -------
    None.
    '''
    
    # Create an append filter to add all the shape polydata to
    append_filter = vtk.vtkAppendPolyData()
    
    # Check if nodes are to be rendered
    if render_nodes == True:
        
        # Add the deformed nodes to the append filter
        for node in model.nodes.values():
            
            vis_node = VisDeformedNode(node, scale_factor, annotation_size, combo_name)
            append_filter.AddInputData(vis_node.source.GetOutput())
        
    # Add the springs to the append filter
    for spring in model.springs.values():
        
        # Only add the spring if it is active for the given load combination
        if spring.active[combo_name] == True:
            
            vis_spring = VisDeformedSpring(spring, model.nodes, scale_factor, combo_name)
            append_filter.AddInputData(vis_spring.source.GetOutput())
            
    # Add the members to the append filter
    for member in model.members.values():
        
        # Only add the member if it is active for the given load combination.
        if member.active[combo_name] == True:
            
            vis_member = VisDeformedMember(member, model.nodes, scale_factor, combo_name)
            append_filter.AddInputData(vis_member.source)
            
    # Create a mapper and actor for the append filter
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(append_filter.GetOutputPort())
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)

    # Adjust the color
    if theme == 'default':
        actor.GetProperty().SetColor(255/255, 255/255, 0/255)  # Yellow
    elif theme == 'print':
        actor.GetProperty().SetColor(26/255, 26/255, 26/255)  # Dark Grey
    
    # Add the actor to the renderer
    vtk_renderer.AddActor(actor)

def _RenderLoads(model, renderer, annotation_size, combo_name, case, theme='default'):

    # Create an append filter to store all the polydata in. This will allow us to use fewer actors to
    # display all the loads, which will greatly improve rendering speed as the user interacts. VTK
    # becomes very slow when a large number of actors are used.
    polydata = vtk.vtkAppendPolyData()
    
    # Polygons are treated as cells in VTK. Create a cell array to store all the area load polygons
    # in. We'll also create a list of points to store the polygon points in. The polydata for these
    # polygons will be stored separately from the other load data.
    polygons = vtk.vtkCellArray()
    polygon_points = vtk.vtkPoints()
    polygon_polydata = vtk.vtkPolyData()
    
    # Get the maximum load magnitudes that will be used to normalize the display scale
    max_pt_load, max_moment, max_dist_load, max_area_load = _MaxLoads(model, combo_name, case)
    
    # Display the requested load combination, or 'Combo 1' if no load combo or case has been
    # specified
    if case == None:
        # Store model.load_combos[combo].factors under a simpler name for use below
        load_factors = model.load_combos[combo_name].factors
    else:
        # Set up a load combination dictionary that represents the load case
        load_factors = {case: 1}
    
    # Step through each node
    for node in model.nodes.values():
    
        # Step through and display each nodal load
        for load in node.NodeLoads:
          
            # Determine if this load is part of the requested LoadCombo or case
            if load[2] in load_factors:
              
                # Calculate the factored value for this load and it's sign (positive or negative)
                load_value = load[1]*load_factors[load[2]]
                if load_value != 0:
                    sign = load_value/abs(load_value)
                else:
                    sign = 1
                
                # Display the load
                if load[0] == 'FX':
                    ptLoad = VisPtLoad((node.X - 0.6*annotation_size*sign, node.Y, node.Z), [1, 0, 0], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'FY':
                    ptLoad = VisPtLoad((node.X, node.Y - 0.6*annotation_size*sign, node.Z), [0, 1, 0], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'FZ':
                    ptLoad = VisPtLoad((node.X, node.Y, node.Z - 0.6*annotation_size*sign), [0, 0, 1], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'MX':
                    ptLoad = VisMoment((node.X, node.Y, node.Z), (1*sign, 0, 0), abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'MY':
                    ptLoad = VisMoment((node.X, node.Y, node.Z), (0, 1*sign, 0), abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'MZ':
                    ptLoad = VisMoment((node.X, node.Y, node.Z), (0, 0, 1*sign), abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                
                polydata.AddInputData(ptLoad.polydata.GetOutput())
                renderer.AddActor(ptLoad.lblActor)
                ptLoad.lblActor.SetCamera(renderer.GetActiveCamera())
    
    # Step through each member
    for member in model.members.values():
    
        # Get the direction cosines for the member's local axes
        dir_cos = member.T()[0:3, 0:3]
      
        # Get the starting point for the member
        x_start, y_start, z_start = member.i_node.X, member.i_node.Y, member.i_node.Z
      
        # Step through each member point load
        for load in member.PtLoads:
      
            # Determine if this load is part of the requested load combination
            if load[3] in load_factors:
        
                # Calculate the factored value for this load and it's sign (positive or negative)
                load_value = load[1]*load_factors[load[3]]
                sign = load_value/abs(load_value)
          
                # Calculate the load's location in 3D space
                x = load[2]
                position = [x_start + dir_cos[0, 0]*x, y_start + dir_cos[0, 1]*x, z_start + dir_cos[0, 2]*x]
          
                # Display the load
                if load[0] == 'Fx':
                    ptLoad = VisPtLoad(position, dir_cos[0, :], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'Fy':
                    ptLoad = VisPtLoad(position, dir_cos[1, :], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'Fz':
                    ptLoad = VisPtLoad(position, dir_cos[2, :], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'Mx':
                    ptLoad = VisMoment(position, dir_cos[0, :]*sign, abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'My':
                    ptLoad = VisMoment(position, dir_cos[1, :]*sign, abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'Mz':
                    ptLoad = VisMoment(position, dir_cos[2, :]*sign, abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'FX':
                    ptLoad = VisPtLoad(position, [1, 0, 0], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'FY':
                    ptLoad = VisPtLoad(position, [0, 1, 0], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'FZ':
                    ptLoad = VisPtLoad(position, [0, 0, 1], load_value/max_pt_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'MX':
                    ptLoad = VisMoment(position, [1*sign, 0, 0], abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'MY':
                    ptLoad = VisMoment(position, [0, 1*sign, 0], abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
                elif load[0] == 'MZ':
                    ptLoad = VisMoment(position, [0, 0, 1*sign], abs(load_value)/max_moment*2.5*annotation_size, '{:.3g}'.format(load_value), annotation_size)
            
                polydata.AddInputData(ptLoad.polydata.GetOutput())
                renderer.AddActor(ptLoad.lblActor)
                ptLoad.lblActor.SetCamera(renderer.GetActiveCamera())
    
        # Step through each member distributed load
        for load in member.DistLoads:
    
            # Determine if this load is part of the requested load combination
            if load[5] in load_factors:
        
                # Calculate the factored value for this load and it's sign (positive or negative)
                w1 = load[1]*load_factors[load[5]]
                w2 = load[2]*load_factors[load[5]]
          
                # Calculate the loads location in 3D space
                x1 = load[3]
                x2 = load[4]
                position1 = [x_start + dir_cos[0, 0]*x1, y_start + dir_cos[0, 1]*x1, z_start + dir_cos[0, 2]*x1]
                position2 = [x_start + dir_cos[0, 0]*x2, y_start + dir_cos[0, 1]*x2, z_start + dir_cos[0, 2]*x2]
                
                # Display the load
                if load[0] == 'Fx':
                    distLoad = VisDistLoad(position1, position2, dir_cos[0, :], w1/max_dist_load*5*annotation_size, w2/max_dist_load*5*annotation_size, '{:.3g}'.format(w1), '{:.3g}'.format(w2), annotation_size)
                elif load[0] == 'Fy':
                    distLoad = VisDistLoad(position1, position2, dir_cos[1, :], w1/max_dist_load*5*annotation_size, w2/max_dist_load*5*annotation_size, '{:.3g}'.format(w1), '{:.3g}'.format(w2), annotation_size)
                elif load[0] == 'Fz':
                    distLoad = VisDistLoad(position1, position2, dir_cos[2, :], w1/max_dist_load*5*annotation_size, w2/max_dist_load*5*annotation_size, '{:.3g}'.format(w1), '{:.3g}'.format(w2), annotation_size)
                elif load[0] == 'FX':
                    distLoad = VisDistLoad(position1, position2, [1, 0, 0], w1/max_dist_load*5*annotation_size, w2/max_dist_load*5*annotation_size, '{:.3g}'.format(w1), '{:.3g}'.format(w2), annotation_size)
                elif load[0] == 'FY':
                    distLoad = VisDistLoad(position1, position2, [0, 1, 0], w1/max_dist_load*5*annotation_size, w2/max_dist_load*5*annotation_size, '{:.3g}'.format(w1), '{:.3g}'.format(w2), annotation_size)
                elif load[0] == 'FZ':
                    distLoad = VisDistLoad(position1, position2, [0, 0, 1], w1/max_dist_load*5*annotation_size, w2/max_dist_load*5*annotation_size, '{:.3g}'.format(w1), '{:.3g}'.format(w2), annotation_size)
               
                polydata.AddInputData(distLoad.polydata.GetOutput())
                renderer.AddActor(distLoad.lblActors[0])
                renderer.AddActor(distLoad.lblActors[1])
                distLoad.lblActors[0].SetCamera(renderer.GetActiveCamera())
                distLoad.lblActors[1].SetCamera(renderer.GetActiveCamera())   
    
    # Step through each plate
    i = 0
    for plate in list(model.plates.values()) + list(model.quads.values()):
    
        # Get the direction cosines for the plate's local z-axis
        dir_cos = plate.T()[0:3, 0:3]
        dir_cos = dir_cos[2]
      
        # Step through each plate load
        for load in plate.pressures:
      
            # Determine if this load is part of the requested load combination
            if load[1] in load_factors:
        
                # Calculate the factored value for this load
                load_value = load[0]*load_factors[load[1]]
                
                # Find the sign for this load. Intercept any divide by zero errors
                if load[0] == 0:
                    sign = 1
                else:
                    sign = abs(load[0])/load[0]
          
                # Find the position of the load's 4 corners
                position0 = [plate.i_node.X, plate.i_node.Y, plate.i_node.Z]
                position1 = [plate.j_node.X, plate.j_node.Y, plate.j_node.Z]
                position2 = [plate.m_node.X, plate.m_node.Y, plate.m_node.Z]
                position3 = [plate.n_node.X, plate.n_node.Y, plate.n_node.Z]
          
                # Create an area load and get its data
                area_load = VisAreaLoad(position0, position1, position2, position3, dir_cos*sign, abs(load_value)/max_area_load*5*annotation_size, '{:.3g}'.format(load_value), annotation_size, theme)
          
                # Add the area load's arrows to the overall load polydata
                polydata.AddInputData(area_load.polydata.GetOutput())
          
                # Add the 4 points at the corners of this area load to the list of points
                polygon_points.InsertNextPoint(area_load.p0[0], area_load.p0[1], area_load.p0[2])
                polygon_points.InsertNextPoint(area_load.p1[0], area_load.p1[1], area_load.p1[2])
                polygon_points.InsertNextPoint(area_load.p2[0], area_load.p2[1], area_load.p2[2])
                polygon_points.InsertNextPoint(area_load.p3[0], area_load.p3[1], area_load.p3[2])
          
                # Create a polygon based on the four points we just defined.
                # The 1st number in `SetId()` is the local point id
                # The 2nd number in `SetId()` is the global point id
                polygon = vtk.vtkPolygon()
                polygon.GetPointIds().SetNumberOfIds(4)
                polygon.GetPointIds().SetId(0, i*4)
                polygon.GetPointIds().SetId(1, i*4 + 1)
                polygon.GetPointIds().SetId(2, i*4 + 2)
                polygon.GetPointIds().SetId(3, i*4 + 3)
          
                # Add the polygon to the list of polygons
                polygons.InsertNextCell(polygon)
                
                # Add the load label
                renderer.AddActor(area_load.label_actor)
          
                # Set the text to follow the camera as the user interacts
                area_load.label_actor.SetCamera(renderer.GetActiveCamera())
          
                # `i` keeps track of the next polygon's ID. We've just added a polygon, so `i` needs to
                # go up 1.
                i += 1
        
                # Create polygon polydata from all the points and polygons we just defined
                polygon_polydata.SetPoints(polygon_points)
                polygon_polydata.SetPolys(polygons)
    
    # Set up an actor and mapper for the loads
    load_mapper = vtk.vtkPolyDataMapper()
    load_mapper.SetInputConnection(polydata.GetOutputPort())
    load_actor = vtk.vtkActor()
    load_actor.SetMapper(load_mapper)

    # Colorize the loads
    if theme == 'default':
        load_actor.GetProperty().SetColor(0/255, 255/255, 0/255)  # Green
    elif theme == 'print':
        load_actor.GetProperty().SetColor(255/255, 0/255, 0/255)  # Red

    # Add the load actor to the renderer
    renderer.AddActor(load_actor)
    
    # Set up an actor and a mapper for the area load polygons
    polygon_mapper = vtk.vtkPolyDataMapper()
    polygon_mapper.SetInputData(polygon_polydata)
    polygon_actor = vtk.vtkActor()

    # polygon_actor.GetProperty().SetOpacity(0.5)      # 50% opacity
    polygon_actor.SetMapper(polygon_mapper)
    renderer.AddActor(polygon_actor)
    
    # Set the color of the area load polygons
    if theme == 'default':
        polygon_actor.GetProperty().SetColor(0/255, 255/255, 0/255)  # Green
    elif theme == 'print':
        polygon_actor.GetProperty().SetColor(255/255, 0/255, 0/255)  # Red

def _RenderContours(model, renderer, deformed_shape, deformed_scale, color_map, scalar_bar, scalar_bar_text_size, combo_name, theme='default'):
  
    # Create a new `vtkCellArray` object to store the elements
    plates = vtk.vtkCellArray()
    
    # Create a `vtkPoints` object to store the coordinates of the corners of the elements
    plate_points = vtk.vtkPoints()
    
    # Create 2 lists to store plate result
    # `results` will store the results in a Python iterable list
    # `plate_results` will store the results in a `vtkDoubleArray` for VTK
    results = []
    plate_results = vtk.vtkDoubleArray()
    plate_results.SetNumberOfComponents(1)
    
    # Each element will be assigned a unique element number `i` beginning at 0
    i = 0
    
    # Calculate the smoothed contour results at each node
    _PrepContour(model, color_map, combo_name)
    
    # Add each plate and quad in the model to the cell array we just created
    for item in list(model.plates.values()) + list(model.quads.values()):
        
        # Create a point for each corner (must be in counter clockwise order)
        if deformed_shape == True:
            p0 = [item.i_node.X + item.i_node.DX[combo_name]*deformed_scale,
                  item.i_node.Y + item.i_node.DY[combo_name]*deformed_scale,
                  item.i_node.Z + item.i_node.DZ[combo_name]*deformed_scale]
            p1 = [item.j_node.X + item.j_node.DX[combo_name]*deformed_scale,
                  item.j_node.Y + item.j_node.DY[combo_name]*deformed_scale,
                  item.j_node.Z + item.j_node.DZ[combo_name]*deformed_scale]
            p2 = [item.m_node.X + item.m_node.DX[combo_name]*deformed_scale,
                  item.m_node.Y + item.m_node.DY[combo_name]*deformed_scale,
                  item.m_node.Z + item.m_node.DZ[combo_name]*deformed_scale]
            p3 = [item.n_node.X + item.n_node.DX[combo_name]*deformed_scale,
                  item.n_node.Y + item.n_node.DY[combo_name]*deformed_scale,
                  item.n_node.Z + item.n_node.DZ[combo_name]*deformed_scale]
        else:
            p0 = [item.i_node.X, item.i_node.Y, item.i_node.Z]
            p1 = [item.j_node.X, item.j_node.Y, item.j_node.Z]
            p2 = [item.m_node.X, item.m_node.Y, item.m_node.Z]
            p3 = [item.n_node.X, item.n_node.Y, item.n_node.Z]
      
        # Add the points to the `vtkPoints` object we created earlier
        plate_points.InsertNextPoint(p0)
        plate_points.InsertNextPoint(p1)
        plate_points.InsertNextPoint(p2)
        plate_points.InsertNextPoint(p3)
      
        # Create a `vtkQuad` based on the four points we just defined
        # The 1st number in `SetId()` is the local point id
        # The 2nd number in `SetId()` is the global point id
        quad = vtk.vtkQuad()
        quad.GetPointIds().SetId(0, i*4)
        quad.GetPointIds().SetId(1, i*4 + 1)
        quad.GetPointIds().SetId(2, i*4 + 2)
        quad.GetPointIds().SetId(3, i*4 + 3)
      
        # Get the contour value for each node
        r0 = item.i_node.contour
        r1 = item.j_node.contour
        r2 = item.m_node.contour
        r3 = item.n_node.contour
            
        if color_map != None:
              
            # Save the results to the Python list of results we created earlier
            results.append(r0)
            results.append(r1)
            results.append(r2)
            results.append(r3)
                  
            # Save the results to the `vtkDoubleArray` list of results for VTK
            plate_results.InsertNextTuple([r0])
            plate_results.InsertNextTuple([r1])
            plate_results.InsertNextTuple([r2])
            plate_results.InsertNextTuple([r3])
      
        # Insert the quad into the cell array
        plates.InsertNextCell(quad)
      
        # Increment `i` for the next plate
        i += 1

    # Create a `vtkPolyData` object to store plate data in
    plate_polydata = vtk.vtkPolyData()
      
    # Add the points and plates to the dataset
    plate_polydata.SetPoints(plate_points)
    plate_polydata.SetPolys(plates)
      
    # Setup actor and mapper for the plates
    plate_mapper = vtk.vtkPolyDataMapper()
    plate_mapper.SetInputData(plate_polydata)
    plate_actor = vtk.vtkActor()
    plate_actor.SetMapper(plate_mapper)
    
    # Map the results to the plates
    if color_map != None:
          
        plate_polydata.GetPointData().SetScalars(plate_results)
            
        # Create a `vtkLookupTable` for the colors used to map results
        lut = vtk.vtkLookupTable()
        lut.SetTableRange(min(results), max(results))
        lut.SetNumberOfColors(256) 
        # The commented code below can be uncommented and modified to change the color scheme
        # ctf = vtk.vtkColorTransferFunction()
        # ctf.SetColorSpaceToDiverging()
        # ctf.AddRGBPoint(min(results), 255, 0, 255)  # Purple
        # ctf.AddRGBPoint(max(results), 255, 0, 0)    # Red
        # for i in range(256):
        #   rgb = list(ctf.GetColor(float(i)/256))
        #   rgb.append(1.0)
        #   lut.SetTableValue(i, *rgb)
        plate_mapper.SetLookupTable(lut)
        plate_mapper.SetUseLookupTableScalarRange(True)
        plate_mapper.SetScalarModeToUsePointData()
        lut.Build()
      
        # Add the scalar bar for the contours.
        if scalar_bar:
            
            if Renderer.scalar == None:
                Renderer.scalar = vtk.vtkScalarBarActor()
            
            scalar = Renderer.scalar

            # This next group of lines controls the font on the scalar bar
            scalar.SetUnconstrainedFontSize(1)
            scalar_text = vtk.vtkTextProperty()
            scalar_text.SetFontSize(max(int(scalar_bar_text_size), 1))
            scalar_text.SetBold(1)

            # The `vtkTextProperty` object is white by default
            if theme == 'print':
                scalar_text.SetColor(255/255, 255/255, 255/255)  # Black
            
            scalar.SetLabelTextProperty(scalar_text)

            scalar.SetMaximumWidthInPixels(100)

            scalar.SetTextPositionToPrecedeScalarBar()

            scalar.SetLookupTable(lut)

            renderer.AddActor(scalar)
      
    # Add the actor for the plates
    renderer.AddActor(plate_actor)

def _MaxLoads(model, combo_name=None, case=None):

    max_pt_load = 0
    max_moment = 0
    max_dist_load = 0
    max_area_load = 0
    
    # Find the requested load combination or load case
    if case == None:
    
        # Step through each node
        for node in model.nodes.values():
      
            # Step through each nodal load to find the largest one
            for load in node.NodeLoads:
              
                # Find the largest loads in the load combination
                if load[2] in model.load_combos[combo_name].factors:
                    if load[0] == 'FX' or load[0] == 'FY' or load[0] == 'FZ':
                        if abs(load[1]*model.load_combos[combo_name].factors[load[2]]) > max_pt_load:
                            max_pt_load = abs(load[1]*model.load_combos[combo_name].factors[load[2]])
                    else:
                        if abs(load[1]*model.load_combos[combo_name].factors[load[2]]) > max_moment:
                            max_moment = abs(load[1]*model.load_combos[combo_name].factors[load[2]])
    
        # Step through each member
        for member in model.members.values():
      
            # Step through each member point load
            for load in member.PtLoads:
                
                # Find and store the largest point load and moment in the load combination
                if load[3] in model.load_combos[combo_name].factors:
          
                    if (load[0] == 'Fx' or load[0] == 'Fy' or load[0] == 'Fz'
                    or  load[0] == 'FX' or load[0] == 'FY' or load[0] == 'FZ'):
                        if abs(load[1]*model.load_combos[combo_name].factors[load[3]]) > max_pt_load:
                            max_pt_load = abs(load[1]*model.load_combos[combo_name].factors[load[3]])
                    else:
                        if abs(load[1]*model.load_combos[combo_name].factors[load[3]]) > max_moment:
                            max_moment = abs(load[1]*model.load_combos[combo_name].factors[load[3]])
      
            # Step through each member distributed load
            for load in member.DistLoads:
        
                #Find and store the largest distributed load in the load combination
                if load[5] in model.load_combos[combo_name].factors:
          
                    if abs(load[1]*model.load_combos[combo_name].factors[load[5]]) > max_dist_load:
                        max_dist_load = abs(load[1]*model.load_combos[combo_name].factors[load[5]])
                    if abs(load[2]*model.load_combos[combo_name].factors[load[5]]) > max_dist_load:
                        max_dist_load = abs(load[2]*model.load_combos[combo_name].factors[load[5]])
      
        # Step through each plate
        for plate in model.plates.values():
      
            # Step through each plate load
            for load in plate.pressures:
        
                if load[1] in model.load_combos[combo_name].factors:
                    if abs(load[0]*model.load_combos[combo_name].factors[load[1]]) > max_area_load:
                        max_area_load = abs(load[0]*model.load_combos[combo_name].factors[load[1]])
      
        # Step through each quad
        for quad in model.quads.values():
      
            # Step through each plate load
            for load in quad.pressures:
        
                # Check to see if the load case is in the requested load combination
                if load[1] in model.load_combos[combo_name].factors:
                    if abs(load[0]*model.load_combos[combo_name].factors[load[1]]) > max_area_load:
                        max_area_load = abs(load[0]*model.load_combos[combo_name].factors[load[1]])
      
    # Behavior if case has been specified
    else:
        
        # Step through each node
        for node in model.nodes.values():
      
            # Step through each nodal load to find the largest one
            for load in node.NodeLoads:
              
                # Find the largest loads in the load case
                if load[2] == case:
                    if load[0] == 'FX' or load[0] == 'FY' or load[0] == 'FZ':
                        if abs(load[1]) > max_pt_load:
                            max_pt_load = abs(load[1])
                    else:
                        if abs(load[1]) > max_moment:
                            max_moment = abs(load[1])
      
        # Step through each member
        for member in model.members.values():
      
            # Step through each member point load
            for load in member.PtLoads:
              
                # Find and store the largest point load and moment in the load case
                if load[3] == case:
            
                    if (load[0] == 'Fx' or load[0] == 'Fy' or load[0] == 'Fz'
                    or  load[0] == 'FX' or load[0] == 'FY' or load[0] == 'FZ'):
                        if abs(load[1]) > max_pt_load:
                            max_pt_load = abs(load[1])
                    else:
                        if abs(load[1]) > max_moment:
                            max_moment = abs(load[1])
        
            # Step through each member distributed load
            for load in member.DistLoads:
          
                # Find and store the largest distributed load in the load case
                if load[5] == case:
            
                    if abs(load[1]) > max_dist_load:
                        max_dist_load = abs(load[1])
                    if abs(load[2]) > max_dist_load:
                        max_dist_load = abs(load[2])
            
            # Step through each plate
            for plate in model.plates.values():
          
                # Step through each plate load
                for load in plate.pressures:
          
                    if load[1] == case:
            
                        if abs(load[0]) > max_area_load:
                            max_area_load = abs(load[0])
          
        # Step through each quad
        for quad in model.quads.values():
      
            # Step through each plate load
            for load in quad.pressures:
        
                if load[1] == case:
          
                    if abs(load[0]) > max_area_load:
                        max_area_load = abs(load[0])
        
    # Return the maximum loads in the load combination or load case
    return max_pt_load, max_moment, max_dist_load, max_area_load



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PyNiteFEA",
    version="1.0.0",
    author="D. Craig Brinck, PE, SE",
    author_email="Building.Code@outlook.com",
    description="A simple elastic 3D structural finite element library for Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JWock82/Pynite.git",
    packages=setuptools.find_packages(include=['Pynite', 'Pynite.*']),
    package_data = {'Pynite': ['*html', '*.css', '*Full Logo No Buffer - Transparent.png']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
        'PrettyTable',
        'scipy',
        'matplotlib',
    ],
    extras_require = {
        'all': ['Ipython', 'vtk', 'pyvista[all,trame]', 'trame_jupyter_extension', 'ipywidgets', 'pdfkit', 'Jinja2'],
        'vtk':  ['IPython', 'vtk'],
        'pyvista': ['pyvista[all,trame]', 'trame_jupyter_extension', 'ipywidgets'],
        'reporting': ['pdfkit', 'Jinja2'],
        'derivations': ['jupyterlab', 'sympy']
    },
    include_package_data = True,
    python_requires = '>=3.7',
)
